from dataclasses import dataclass

from auto_editor.utils.types import (
    float_type, anchor_type, color_type, text_content, align_type
)

@dataclass
class TimelineObject:
    start: int
    dur: int

@dataclass
class RectangleObject(TimelineObject):
    x: int
    y: int
    width: int
    height: int
    anchor: anchor_type = 'ce'
    fill: color_type = '#c4c4c4'
    stroke: int = 0
    strokecolor: color_type = '#000'
    _type: str = 'rectangle'

@dataclass
class EllipseObject(TimelineObject):
    x: int
    y: int
    width: int
    height: int
    anchor: anchor_type = 'ce'
    fill: color_type = '#c4c4c4'
    stroke: int = 0
    strokecolor: color_type = '#000'
    _type: str = 'ellipse'

@dataclass
class TextObject(TimelineObject):
    content: text_content
    x: int = 'centerX'
    y: int = 'centerY'
    size: int = 30
    font: str = 'default'
    align: align_type = 'left'
    fill: color_type = '#000'
    stroke: int = 0
    strokecolor: color_type = '#000'
    _type: str = 'text'

@dataclass
class ImageObject(TimelineObject):
    src: str
    x: int = 'centerX'
    y: int = 'centerY'
    opacity: float_type = 1
    anchor: anchor_type = 'ce'
    rotate: float = 0 # in degrees
    _type: str = 'image'



"""
Make a pre-edited file reference that can be inputted back into auto-editor.
"""

import os
import sys
import json

from auto_editor.utils.log import Log

def read_json_timeline(json_file: str, log: Log):
    with open(json_file, 'r') as f:
        data = json.load(f)

    source = data['source']
    if(not os.path.isfile(source)):
        log.error(f"Could not locate media file: '{source}'")

    background = data['background']

    return background, source, data['chunks']


def make_json_timeline(media_file, out, obj_sheet, chunks, fps, background, log):
    data = {
        'version': '0.2.0',
        'source': os.path.abspath(media_file),
        'background': background,
        'chunks': chunks,
        'timeline': obj_sheet.all,
    }

    if(isinstance(out, str)):
        if(not out.endswith('.json')):
            log.error('Output extension must be .json')

        with open(out, 'w') as outfile:
            json.dump(data, outfile, indent=2, default=lambda o: o.__dict__)
    else:
        json.dump(data, sys.stdout, indent=2, default=lambda o: o.__dict__)




import sys

from math import floor
from time import time, localtime
from shutil import get_terminal_size
from platform import system

from .func import get_stdout


class ProgressBar:
    def __init__(self, bar_type: str):

        self.machine = False
        self.hide = False

        self.icon = '⏳'
        self.chars = (' ', '▏', '▎', '▍', '▌', '▋', '▊', '▉', '█')
        self.brackets = ('|', '|')

        if bar_type == 'classic':
            self.icon = '⏳'
            self.chars = ('░', '█')
            self.brackets = ('[', ']')
        if bar_type == 'ascii':
            self.icon = '& '
            self.chars = ('-', '#')
            self.brackets = ('[', ']')
        if bar_type == 'machine':
            self.machine = True
        if bar_type == 'none':
            self.hide = True

        self.part_width = len(self.chars) - 1

        self.ampm = True
        if system() == 'Darwin' and bar_type in ('default', 'classic'):
            try:
                date_format = get_stdout(
                    ['defaults', 'read', 'com.apple.menuextra.clock', 'DateFormat']
                )
                self.ampm = 'a' in date_format
            except FileNotFoundError:
                pass


    @staticmethod
    def pretty_time(my_time: float, ampm: bool) -> str:
        new_time = localtime(my_time)

        hours = new_time.tm_hour
        minutes = new_time.tm_min

        if ampm:
            if hours == 0:
                hours = 12
            if hours > 12:
                hours -= 12
            ampm_marker = 'PM' if new_time.tm_hour >= 12 else 'AM'
            return '{:02}:{:02} {}'.format(hours, minutes, ampm_marker)
        return '{:02}:{:02}'.format(hours, minutes)

    def tick(self, index):

        if self.hide:
            return

        progress = min(1, max(0, index / self.total))

        if progress == 0:
            progress_rate = 0
        else:
            progress_rate = (time() - self.begin_time) / progress

        if self.machine:
            index = min(index, self.total)
            raw = int(self.begin_time + progress_rate)
            print('{}~{}~{}~{}~{}'.format(
                self.title, index, self.total, self.begin_time, raw),
                end='\r', flush=True)
            return


        new_time = self.pretty_time(self.begin_time + progress_rate, self.ampm)

        percent = round(progress * 100, 1)
        p_pad = " " * (4 - len(str(percent)))

        columns = get_terminal_size().columns
        bar_len = max(1, columns - (self.len_title + 32))
        done_nums = round(percent * bar_len)

        progress_bar_str = self.progress_bar_str(progress, bar_len)

        bar = f'  {self.icon}{self.title} {progress_bar_str} {p_pad}{percent}%  ETA {new_time}'

        if len(bar) > columns - 2:
            bar = bar[:columns - 2]
        else:
            bar += ' ' * (columns - len(bar) - 4)

        sys.stdout.write(bar + '\r')
        try:
            sys.stdout.flush()
        except AttributeError:
            pass

    def start(self, total, title: str='Please wait'):
        self.title = title
        self.len_title = len(title)
        self.total = total
        self.begin_time = time()

        try:
            self.tick(0)
        except UnicodeEncodeError:
            self.icon = '& '
            self.chars = ('-', '#')
            self.part_width = 1

    def progress_bar_str(self, progress: float, width: int) -> str:
        whole_width = floor(progress * width)
        remainder_width = (progress * width) % 1
        part_width = floor(remainder_width * self.part_width)
        part_char = self.chars[part_width]

        if width - whole_width - 1 < 0:
            part_char = ''

        line = (
            self.brackets[0]
            + self.chars[-1] * whole_width
            + part_char
            + self.chars[0] * (width - whole_width - 1)
            + self.brackets[1]
        )
        return line

    @staticmethod
    def end():
        sys.stdout.write(' ' * (get_terminal_size().columns - 2) + '\r')



import os
import re
import sys
from typing import List, Tuple, Optional

from auto_editor.utils.progressbar import ProgressBar
from auto_editor.utils.log import Log
from auto_editor.ffwrapper import FFmpeg


class MyLogger:
    @staticmethod
    def debug(msg):
        pass

    @staticmethod
    def warning(msg):
        print(msg, file=sys.stderr)

    @staticmethod
    def error(msg):
        if "'Connection refused'" in msg:
            pass
        else:
            print(msg, file=sys.stderr)


def parse_bytes(bytestr) -> Optional[int]:
    # Parse a string indicating a byte quantity into an integer.
    matchobj = re.match(r'(?i)^(\d+(?:\.\d+)?)([kMGTPEZY]?)$', bytestr)
    if matchobj is None:
        return None
    number = float(matchobj.group(1))
    multiplier = 1024.0 ** 'bkmgtpezy'.index(matchobj.group(2).lower())
    return round(number * multiplier)


def sponsor_block_api(_id: str, categories: List[str], log: Log) -> Optional[dict]:
    from urllib import request
    from urllib.error import HTTPError
    import json

    cat_url = 'categories=["' + '","'.join(categories) + '"]'

    try:
        contents = request.urlopen(
            f'https://sponsor.ajay.app/api/skipSegments?videoID={_id}&{cat_url}'
        )

        return json.loads(contents.read())
    except HTTPError:
        log.warning("Couldn't find skipSegments for id: {}".format(_id))
        return None


def download_video(my_input: str, args, ffmpeg: FFmpeg, log: Log) -> str:
    log.conwrite('Downloading video...')
    if ' @' in my_input:
        max_height = my_input[my_input.index(' @') + 2 :].strip()
        download_format = f'bestvideo[ext=mp4][height<={max_height}]+bestaudio[ext=m4a]'

        my_input = my_input[: my_input.index(' ')]
    else:
        download_format = 'bestvideo[ext=mp4]+bestaudio[ext=m4a]'

    outtmpl = re.sub(r'\W+', '-', os.path.splitext(my_input)[0]) + '.mp4'

    if args.download_dir is not None:
        outtmpl = os.path.join(args.download_dir, outtmpl)

    try:
        import yt_dlp
    except ImportError:
        log.import_error('yt-dlp')

    if not os.path.isfile(outtmpl):
        ytbar = ProgressBar(args.progress)
        ytbar.start(100, 'Downloading')

        def my_hook(d: dict) -> None:
            if d['status'] == 'downloading':
                ytbar.tick(float(d['_percent_str'].replace('%', '')))

        def abspath(path: Optional[str]) -> Optional[str]:
            if path is None:
                return None
            return os.path.abspath(path)

        ydl_opts = {
            'outtmpl': outtmpl,
            'ffmpeg_location': ffmpeg.path,
            'format': download_format,
            'ratelimit': parse_bytes(args.limit_rate),
            'logger': MyLogger(),
            'cookiefile': abspath(args.cookies),
            'download_archive': abspath(args.download_archive),
            'progress_hooks': [my_hook],
        }

        for item, key in ydl_opts.items():
            if item is None:
                del ydl_opts[key]

        try:
            with yt_dlp.YoutubeDL(ydl_opts) as ydl:
                ydl.download([my_input])
        except yt_dlp.utils.DownloadError as error:
            if 'format is not available' in str(error):
                del ydl_opts['format']
                with yt_dlp.YoutubeDL(ydl_opts) as ydl:
                    ydl.download([my_input])
            else:
                log.error('yt-dlp: Download Error.')

        log.conwrite('')
    return outtmpl


def get_segment(args, my_input: str, log: Log) -> Optional[dict]:
    if args.block is not None:
        if args.id is not None:
            return sponsor_block_api(args.id, args.block, log)
        match = re.search(
            r'youtube\.com/watch\?v=(?P<match>[A-Za-z0-9_-]{11})', my_input
        )
        if match:
            youtube_id = match.groupdict()['match']
            return sponsor_block_api(youtube_id, args.block, log)
    return None


def valid_input(
    inputs: List[str], ffmpeg: FFmpeg, args, log: Log
    ) -> Tuple[List[str], List[Optional[dict]]]:

    new_inputs = []
    segments = []
    for my_input in inputs:
        if os.path.isfile(my_input):
            _, ext = os.path.splitext(my_input)
            if ext == '':
                log.error('File must have an extension.')

            new_inputs.append(my_input)
            segments.append(get_segment(args, my_input, log))

        elif my_input.startswith('http://') or my_input.startswith('https://'):
            new_inputs.append(download_video(my_input, args, ffmpeg, log))
            segments.append(get_segment(args, my_input, log))
        else:
            if os.path.isdir(my_input):
                log.error('Input must be a file or url.')
            log.error('Could not find file: {}'.format(my_input))

    return new_inputs, segments



def safe_mkdir(path: str) -> str:
    from shutil import rmtree
    from os import mkdir
    try:
        mkdir(path)
    except OSError:
        rmtree(path)
        mkdir(path)
    return path

def get_width_height(inp):
    if len(inp.video_streams) == 0:
        return None, None
    else:
        return inp.video_streams[0]['width'], inp.video_streams[0]['height']

def indent(base: int, *lines: str) -> str:
    new_lines = ''
    for line in lines:
        new_lines += ('\t' * base) + line + '\n'
    return new_lines



# Internal Libraries
import re
import sys
import os.path
from subprocess import Popen, call, PIPE
from platform import system

# Typing
from typing import List, Optional, Any

# Included Libraries
from auto_editor.utils.func import get_stdout
from auto_editor.utils.log import Log

def regex_match(regex: str, text: str) -> Optional[str]:
    match = re.search(regex, text)
    if match:
        return match.groupdict()['match']
    return None


class FFmpeg:
    __slots__ = ('debug', 'path', 'version')

    @staticmethod
    def _set_ff_path(ff_location: Optional[str], my_ffmpeg: bool) -> str:
        if ff_location is not None:
            return ff_location
        if my_ffmpeg or system() not in ('Windows', 'Darwin'):
            return 'ffmpeg'
        program = 'ffmpeg' if system() == 'Darwin' else 'ffmpeg.exe'
        dirpath = os.path.dirname(os.path.realpath(__file__))
        return os.path.join(dirpath, 'ffmpeg', system(), program)

    def __init__(
        self,
        ff_location: Optional[str]=None,
        my_ffmpeg: bool=False,
        debug: bool=False,
    ) -> None:

        self.debug = debug
        self.path = self._set_ff_path(ff_location, my_ffmpeg)
        try:
            _version = get_stdout([self.path, '-version']).split('\n')[0]
            _version = _version.replace('ffmpeg version', '').strip()
            self.version = _version.split(' ')[0]
        except FileNotFoundError:
            if system() == 'Darwin':
                Log().error('No ffmpeg found, download via homebrew or restore the '
                    'included binary.')
            if system() == 'Windows':
                Log().error('No ffmpeg found, download ffmpeg with your favorite package '
                    'manager (ex chocolatey), or restore the included binary.')

            Log().error('ffmpeg must be installed and on PATH.')

    def print(self, message: str) -> None:
        if self.debug:
            print(f'FFmpeg: {message}', file=sys.stderr)

    def print_cmd(self, cmd: List[str]) -> None:
        if self.debug:
            print('FFmpeg run: {}\n'.format(' '.join(cmd)), file=sys.stderr)

    def run(self, cmd: List[str]) -> None:
        cmd = [self.path, '-y', '-hide_banner'] + cmd
        if not self.debug:
            cmd.extend(['-nostats', '-loglevel', 'error'])
        self.print_cmd(cmd)
        call(cmd)

    def run_check_errors(self, cmd: List[str], log: Log, show_out: bool=False) -> None:

        def _run(cmd: List[str]) -> Any:
            process = self.Popen(cmd, stdin=PIPE, stdout=PIPE, stderr=PIPE)

            _, stderr = process.communicate()
            process.stdin.close()
            return stderr.decode()

        output = _run(cmd)

        if 'Try -allow_sw 1' in output:
            cmd.insert(-1, '-allow_sw')
            cmd.insert(-1, '1')
            output = _run(cmd)

        error_list = [
            r"Unknown encoder '.*'",
            r"-q:v qscale not available for encoder\. Use -b:v bitrate instead\.",
            r'Specified sample rate .* is not supported',
            r'Unable to parse option value ".*"',
            r'Error setting option .* to value .*\.',
            r"Undefined constant or missing '.*' in '.*'",
            r'DLL .* failed to open',
            r"Incompatible pixel format '.*' for codec '[A-Za-z0-9_]*'",
        ]

        if self.debug:
            print(f'stderr: {output}')

        for item in error_list:
            check = re.search(item, output)
            if check:
                log.error(check.group())

        if show_out and not self.debug:
            print(f'stderr: {output}')

    def file_info(self, path):
        # type: (str) -> File
        return File(self, path)

    def Popen(self, cmd: List[str], stdin=None, stdout=PIPE, stderr=None):
        cmd = [self.path] + cmd
        self.print_cmd(cmd)
        return Popen(cmd, stdin=stdin, stdout=stdout, stderr=stderr)

    def pipe(self, cmd: List[str]) -> str:
        cmd = [self.path, '-y'] + cmd

        self.print_cmd(cmd)
        output = get_stdout(cmd)
        self.print(output)
        return output


class File:
    __slots__ = ('path', 'abspath', 'basename', 'dirname', 'name', 'ext', 'duration',
        'bitrate', 'metadata', 'fps', 'video_streams', 'audio_streams', 'subtitle_streams')

    def __init__(self, ffmpeg: FFmpeg, path: str):
        self.path = path
        self.abspath = os.path.abspath(path)
        self.basename = os.path.basename(path)
        self.dirname = os.path.dirname(os.path.abspath(path))
        self.name, self.ext = os.path.splitext(path)

        info = get_stdout([ffmpeg.path, '-hide_banner', '-i', path])

        self.duration = regex_match(r'Duration:\s(?P<match>[\d:.]+),', info)
        self.bitrate = regex_match(r'bitrate:\s(?P<match>\d+\skb\/s)', info)

        self.metadata = {}
        active = False
        active_key = None

        for line in info.split('\n'):
            if active:
                if re.search(r'^\s*[A-Z][a-z_]*', line):
                    break

                key = regex_match(r'^\s*(?P<match>[a-z_]+)', line)
                body = regex_match(r'^\s*[a-z_]*\s*:\s(?P<match>[\w\W]*)', line)

                if body is not None:
                    if key is None:
                        if(active_key is not None):
                            self.metadata[active_key] += '\n' + body
                    else:
                        self.metadata[key] = body
                        active_key = key

            if re.search(r'^\s\sMetadata:', line):
                active = True

        video_streams = []
        audio_streams = []
        subtitle_streams = []
        fps = None

        lang_regex = r'Stream #\d+:\d(?:[\[\]a-z0-9]|)+\((?P<match>\w+)\)'

        sub_exts = {'mov_text': 'srt', 'ass': 'ass', 'webvtt': 'vtt'}

        for line in info.split('\n'):
            if re.search(r'Stream #', line):
                s_data = {}
                if re.search(r'Video:', line):
                    s_data['width'] = regex_match(r'(?P<match>\d+)x\d+[\s,]', line)
                    s_data['height'] = regex_match(r'\d+x(?P<match>\d+)[\s,]', line)
                    s_data['codec'] = regex_match(r'Video:\s(?P<match>\w+)', line)
                    s_data['bitrate'] = regex_match(r'\s(?P<match>\d+\skb\/s)', line)
                    s_data['fps'] = regex_match(r'\s(?P<match>[\d\.]+)\stbr', line)
                    s_data['lang'] = regex_match(lang_regex, line)

                    if fps is None:
                        fps = s_data['fps']
                    video_streams.append(s_data)

                elif re.search(r'Audio:', line):
                    s_data['codec'] = regex_match(r'Audio:\s(?P<match>\w+)', line)
                    s_data['samplerate'] = regex_match(r'(?P<match>\d+)\sHz', line)
                    s_data['bitrate'] = regex_match(r'\s(?P<match>\d+\skb\/s)', line)
                    s_data['lang'] = regex_match(lang_regex, line)
                    audio_streams.append(s_data)

                elif re.search(r'Subtitle:', line):
                    s_data['lang'] = regex_match(lang_regex, line)
                    s_data['codec'] = regex_match(r'Subtitle:\s(?P<match>\w+)', line)

                    if s_data['codec'] is not None:
                        s_data['ext'] = sub_exts.get(s_data['codec'], 'vtt')
                    subtitle_streams.append(s_data)

        self.fps = fps
        self.video_streams = video_streams
        self.audio_streams = audio_streams
        self.subtitle_streams = subtitle_streams



#!/usr/bin/env python3

# Internal Libraries
import os
import sys
import tempfile

# Included Libraries
import auto_editor
import auto_editor.vanparse as vanparse
import auto_editor.utils.func as usefulfunctions

from auto_editor.utils.progressbar import ProgressBar
from auto_editor.utils.func import set_output_name
from auto_editor.utils.log import Log, Timer
from auto_editor.ffwrapper import FFmpeg
from auto_editor.edit import edit_media

def main_options(parser):
    from auto_editor.objects import (TextObject, RectangleObject, EllipseObject,
        ImageObject)
    from auto_editor.utils.types import (file_type, float_type, sample_rate_type,
        frame_type, range_type, speed_range_type, block_type, margin_type, color_type)

    parser.add_text('Object Options')
    parser.add_argument('--add-text', nargs='*', dataclass=TextObject,
        help='Add a text object to the timeline.')
    parser.add_argument('--add-rectangle', nargs='*', dataclass=RectangleObject,
        help='Add a rectangle object to the timeline.')
    parser.add_argument('--add-ellipse', nargs='*', dataclass=EllipseObject,
        help='Add an ellipse object to the timeline.',
        manual='The x and y coordinates specify a bounding box where the ellipse is '
            'drawn.')
    parser.add_argument('--add-image', nargs='*', dataclass=ImageObject,
        help='Add an image object onto the timeline.',
        manual='Opacity is how transparent or solid the image is. A transparency of '
            '1 or 100% is completely solid. A transparency of 0 or 0% is completely '
            'transparent.\n'
            'The anchor point tells how the image is placed relative to its x y coordinates.')

    parser.add_text('URL Download Options')
    parser.add_argument('--download-dir', type=str, default=None,
        help='The directory where the downloaded URL is placed.')
    parser.add_argument('--limit-rate', '-rate', default='3m',
        help='The maximum download rate in bytes per second (50k, 4.2m)')
    parser.add_argument('--id', type=str, default=None,
        help='Manually set the YouTube ID the video belongs to.')
    parser.add_argument('--block', type=block_type,
        help='Given a SponsorBlock category, mark the section as silent.',
        manual='Multiple categories can be specified by using a comma: "sponsor,intro"\n'
            'This option only works on YouTube URLs, or where `--id` points to a valid '
            'YouTube ID.\n\nCategories include: '
            'sponsor intro outro selfpromo interaction music_offtopic')
    parser.add_argument('--download-archive', type=file_type, default=None,
        help='Download only videos not listed in archive file. Record the IDs of'
             ' all downloads.')
    parser.add_argument('--cookies', type=file_type, default=None,
        help='The file to read cookies from and dump the cookie jar in.')

    parser.add_text('Motion Detection Options')
    parser.add_argument('--motion-threshold', type=float_type, default=0.02,
        range='0 to 1',
        help='How much motion is required to be considered "moving"')
    parser.add_blank()
    parser.add_argument('--md-width', type=int, default=400, range='1 to Infinity',
        help="Scale the frame to this width before being compared.")
    parser.add_argument('--md-blur', type=int, default=9, range='0 to Infinity',
        help='Set the strength of the blur applied to a frame before being compared.')

    parser.add_text('Exporting as Media Options')
    parser.add_argument('--video-codec', '-vcodec', '-c:v', default='auto',
        help='Set the video codec for the output media file.')
    parser.add_argument('--audio-codec', '-acodec', '-c:a', default='auto',
        help='Set the audio codec for the output media file.')
    parser.add_argument('--video-bitrate', '-b:v', default='10m',
        help='Set the number of bits per second for video.')
    parser.add_argument('--audio-bitrate', '-b:a', default='unset',
        help='Set the number of bits per second for audio.')
    parser.add_argument('--sample-rate', '-ar', type=sample_rate_type,
        help='Set the sample rate of the input and output videos.')
    parser.add_argument('--preset', '-preset', default='unset',
        choices=['ultrafast', 'superfast', 'veryfast', 'faster', 'fast', 'medium',
            'slow', 'slower', 'veryslow', 'unset'],
        help='Set the preset for ffmpeg to help save file size or increase quality.')
    parser.add_argument('--tune', '-tune', default='unset',
        choices=['film', 'animation', 'grain', 'stillimage', 'fastdecode',
            'zerolatency', 'none', 'unset'],
        help='Set the tune for ffmpeg to compress video better in certain circumstances.')
    parser.add_argument('--constant-rate-factor', '-crf', default='unset', range='0 to 51',
        help='Set a value to the ffmpeg option -crf.',
        manual="This option's value gets directly passed to the -crf flag used by ffmpeg "
        "during encoding. If the video encoder does not support crf, "
        "the program will error.")
    parser.add_argument('--video-quality-scale', '-qscale:v', '-q:v', default='unset',
        range='1 to 31',
        help='Set a value to the ffmpeg option -qscale:v')
    parser.add_argument('--scale', type=float_type, default=1,
        help='Scale the output video by a certain factor.')

    parser.add_text('Miscellaneous Options')
    parser.add_argument('--background', type=color_type, default='#000',
        help='Set the color of the background that is visible when the video is moved.')
    parser.add_argument('--combine-files', action='store_true',
        help='Combine all input files into one before editing.')
    parser.add_argument('--progress', default='modern',
        choices=['modern', 'classic', 'ascii', 'machine', 'none'],
        help='Set what type of progress bar to use.')

    parser.add_text('Manual Editing Options')
    parser.add_argument('--cut-out', type=range_type, nargs='*',
        help='The range of media that will be removed completely, regardless of the '
            'value of silent speed.')
    parser.add_argument('--add-in', type=range_type, nargs='*',
        help='The range of media that will be added in, opposite of --cut-out')
    parser.add_blank()
    parser.add_argument('--mark-as-loud', type=range_type, nargs='*',
        help='The range that will be marked as "loud".')
    parser.add_argument('--mark-as-silent', type=range_type, nargs='*',
        help='The range that will be marked as "silent".')
    parser.add_argument('--set-speed-for-range', type=speed_range_type, nargs='*',
        help='Set an arbitrary speed for a given range.',
        manual='This option takes 3 arguments delimited with commas and they are as follows:\n'
            ' Speed\n'
            ' - How fast the media plays. Speeds 0 or below and 99999 or above will be cut completely.\n'
            ' Start\n'
            ' - When the speed first gets applied. The default unit is in frames, but second units can also be used.\n'
            ' End\n'
            ' - When the speed stops being applied. It can use both frame and second units.\n')

    parser.add_text('Select Editing Source Options')
    parser.add_argument('--edit-based-on', '--edit', default='audio',
        choices=['audio', 'motion', 'none', 'all', 'not_audio', 'not_motion',
            'audio_or_motion', 'audio_and_motion', 'audio_xor_motion',
            'audio_and_not_motion', 'not_audio_and_motion', 'not_audio_and_not_motion'],
        help='Decide which method to use when making edits.')
    parser.add_argument('--keep-tracks-seperate', action='store_true',
        help="Don't combine audio tracks when exporting.")
    parser.add_blank()
    parser.add_argument('--cut-by-this-audio', '-ca', type=file_type,
        help="Base cuts by this audio file instead of the video's audio.")
    parser.add_argument('--cut-by-this-track', '-ct', type=int, default=0,
        range='0 to the number of audio tracks minus one',
        help='Base cuts by a different audio track in the video.')
    parser.add_argument('--cut-by-all-tracks', '-cat', action='store_true',
        help='Combine all audio tracks into one before basing cuts.')

    parser.add_text('Export Mode Options')
    parser.add_argument('--export-to-premiere', '-exp', action='store_true',
        help='Export as an XML file for Adobe Premiere Pro instead of making a media file.')
    parser.add_argument('--export-to-final-cut-pro', '-exf', action='store_true',
        help='Export as an XML file for Final Cut Pro instead of making a media file.')
    parser.add_argument('--export-to-shotcut', '-exs', action='store_true',
        help='Export as an XML timeline file for Shotcut instead of making a media file.')
    parser.add_argument('--export-as-json', action='store_true',
        help='Export as a JSON file that can be read by auto-editor later.')
    parser.add_argument('--export-as-audio', '-exa', action='store_true',
        help='Export as a WAV audio file.')
    parser.add_argument('--export-as-clip-sequence', '-excs', action='store_true',
        help='Export as multiple numbered media files.')
    parser.add_argument('--timeline', action='store_true',
        help='Display timeline JSON file and halt.',
        manual='This option is like `--export-as-json` except that it outputs directly '
            'to stdout instead of to a file.')

    parser.add_text('Utility Options')
    parser.add_argument('--no-open', action='store_true',
        help='Do not open the file after editing is done.')
    parser.add_argument('--temp-dir', default=None,
        help='Set where the temporary directory is located.',
        manual='If not set, tempdir will be set with Python\'s tempfile module\n'
            'The directory doesn\'t have to exist beforehand, however, the root path '
            'must be valid. For Windows users, this file will be in the C drive.\n'
            'The temp file can get quite big if you\'re generating a huge video, so '
            'make sure your location has enough space.')
    parser.add_argument('--ffmpeg-location', default=None,
        help='Set a custom path to the ffmpeg location.',
        manual='This takes precedence over `--my-ffmpeg`.')
    parser.add_argument('--my-ffmpeg', action='store_true',
        help='Use the ffmpeg on your PATH instead of the one packaged.',
        manual='This is equivalent to `--ffmpeg-location ffmpeg`.')

    parser.add_text('Display Options')
    parser.add_argument('--version', action='store_true',
        help="Display the program's version and halt.")
    parser.add_argument('--debug', action='store_true',
        help='Show debugging messages and values.')
    parser.add_argument('--show-ffmpeg-debug', action='store_true',
        help='Show ffmpeg progress and output.')
    parser.add_argument('--quiet', '-q', action='store_true',
        help='Display less output.')
    parser.add_argument('--preview', action='store_true',
        help='Show stats on how the input will be cut and halt.')

    parser.add_text('Editing Options')

    parser.add_argument('--silent-threshold', '-t', type=float_type, default=0.04,
        range='0 to 1',
        help='Set the volume that frames audio needs to surpass to be "loud".',
        manual='Silent threshold is a percentage where 0% represents absolute silence and '
            '100% represents the highest volume in the media file.\n'
            'Setting the threshold to `0%` will cut only out areas where area is '
            'absolutely silence while a value of 4% will cut ')
    parser.add_argument('--frame-margin', '--margin', '-m', type=margin_type, default='6',
        range='-Infinity to Infinity',
        help='Set how many "silent" frames on either side on the "loud" sections to be '
            'included.',
        manual='Margin is measured in frames, however, seconds can be used. e.g. `0.3secs`\n'
            'The starting and ending margins can be set separately with the use of '
            'a comma. e.g. `2sec,3sec` `7,10` `-1,6`')
    parser.add_argument('--silent-speed', '-s', type=float_type, default=99999,
        range='Any number. Values <= 0 or >= 99999 will be cut out.',
        help='Set the speed that "silent" sections should be played at.')
    parser.add_argument('--video-speed', '--sounded-speed', '-v', type=float_type,
        default=1,
        range='Any number. Values <= 0 or >= 99999 will be cut out.',
        help='Set the speed that "loud" sections should be played at.')
    parser.add_argument('--min-clip-length', '-mclip', type=frame_type, default=3,
        range='0 to Infinity',
        help='Set the minimum length a clip can be. If a clip is too short, cut it.')
    parser.add_argument('--min-cut-length', '-mcut', type=frame_type, default=6,
        range='0 to Infinity',
        help="Set the minimum length a cut can be. If a cut is too short, don't cut.")

    parser.add_blank()
    parser.add_argument('--help', '-h', action='store_true',
        help='Print info about the program or an option and exit.')

    parser.add_argument('--output-file', '--output', '-o', nargs='*',
        help='Set the name(s) of the new output.')
    parser.add_blank()
    parser.add_required('input', nargs='*',
        help='File(s) or URL(s) that will be edited.')

    return parser


def main():
    desc = ('Auto-Editor is an automatic video/audio creator and editor. '
        'By default, it will detect silence and create a new video with those '
        'sections cut out. By changing some of the options, you can export to a '
        'traditional editor like Premiere Pro and adjust the edits there, adjust '
        'the pacing of the cuts, and change the method of editing like using audio '
        'loudness and video motion to judge making cuts.\nRun:\n    auto-editor '
        '--help\n\nTo get the list of options.\n')

    parser = vanparse.ArgumentParser('Auto-Editor', auto_editor.version, description=desc)

    subcommands = ['create', 'test', 'info', 'levels', 'grep', 'subdump', 'desc']

    if len(sys.argv) > 1 and sys.argv[1] in subcommands:
        obj = __import__('auto_editor.subcommands.{}'.format(sys.argv[1]),
            fromlist=['subcommands'])
        obj.main(sys.argv[2:])
        sys.exit()
    else:
        parser = main_options(parser)
        try:
            args = parser.parse_args(sys.argv[1:])
        except vanparse.ParserError as e:
            Log().error(str(e))

    timer = Timer(args.quiet)

    exporting_to_editor = (args.export_to_premiere or args.export_to_final_cut_pro or
        args.export_to_shotcut)
    making_data_file = exporting_to_editor or args.export_as_json

    is64bit = '64-bit' if sys.maxsize > 2**32 else '32-bit'

    ffmpeg = FFmpeg(args.ffmpeg_location, args.my_ffmpeg, args.show_ffmpeg_debug)

    if args.debug and args.input == []:
        import platform

        print(f'Python Version: {platform.python_version()} {is64bit}')
        print(f'Platform: {platform.system()} {platform.release()} {platform.machine().lower()}')
        print(f'FFmpeg Version: {ffmpeg.version}')
        print(f'FFmpeg Path: {ffmpeg.path}')
        print(f'Auto-Editor Version: {auto_editor.version}')
        sys.exit()

    if is64bit == '32-bit':
        Log().warning('You have the 32-bit version of Python, which may lead to '
            'memory crashes.')

    if args.version:
        print(f'Auto-Editor version {auto_editor.version}')
        sys.exit()

    if args.temp_dir is None:
        TEMP = tempfile.mkdtemp()
    else:
        TEMP = args.temp_dir
        if os.path.isfile(TEMP):
            Log().error('Temp directory cannot be an already existing file.')
        if os.path.isdir(TEMP):
            if len(os.listdir(TEMP)) != 0:
                Log().error('Temp directory should be empty!')
        else:
            os.mkdir(TEMP)

    if args.timeline:
        args.quiet = True

    log = Log(args.debug, args.quiet, temp=TEMP)
    log.debug(f'Temp Directory: {TEMP}')

    if args.input == []:
        log.error('You need to give auto-editor an input file or folder so it can '
            'do the work for you.')

    if([args.export_to_premiere, args.export_to_final_cut_pro, args.export_as_audio,
        args.export_to_shotcut, args.export_as_clip_sequence].count(True) > 1):
        log.error('You must choose only one export option.')

    if args.constant_rate_factor != 'unset':
        if int(args.constant_rate_factor) < 0 or int(args.constant_rate_factor) > 51:
            log.error('Constant rate factor (crf) must be between 0-51.')
    if args.md_width < 1:
        log.error('--md-width cannot be less than 1.')

    def write_starting_message(args) -> str:
        if args.export_to_premiere:
            return 'Exporting to Adobe Premiere Pro XML file'
        if args.export_to_final_cut_pro:
            return 'Exporting to Final Cut Pro XML file'
        if args.export_to_shotcut:
            return 'Exporting to Shotcut XML Timeline file'
        if args.export_as_audio:
            return 'Exporting as audio'
        return 'Starting'

    if not args.preview and not args.timeline:
        log.conwrite(write_starting_message(args))

    if args.preview or args.timeline or args.export_as_clip_sequence or making_data_file:
        args.no_open = True

    if args.md_blur < 0:
        args.md_blur = 0

    if args.silent_speed <= 0 or args.silent_speed > 99999:
        args.silent_speed = 99999

    if args.video_speed <= 0 or args.video_speed > 99999:
        args.video_speed = 99999

    if args.output_file is None:
        args.output_file = []

    from auto_editor.validate_input import valid_input
    input_list, segments = valid_input(args.input, ffmpeg, args, log)

    if len(args.output_file) < len(input_list):
        for i in range(len(input_list) - len(args.output_file)):
            args.output_file.append(set_output_name(input_list[i], None,
                making_data_file, args))

    if args.combine_files:
        if exporting_to_editor:
            temp_file = 'combined.mp4'
        else:
            temp_file = os.path.join(TEMP, 'combined.mp4')

        cmd = []
        for fileref in input_list:
            cmd.extend(['-i', fileref])
        cmd.extend(['-filter_complex', '[0:v]concat=n={}:v=1:a=1'.format(len(input_list)),
            '-codec:v', 'h264', '-pix_fmt', 'yuv420p', '-strict', '-2', temp_file])
        ffmpeg.run(cmd)
        del cmd
        input_list = [temp_file]

    def main_loop(input_list, ffmpeg, args, segments, log):
        num_cuts = 0

        progress = ProgressBar(args.progress)

        for i, input_path in enumerate(input_list):
            inp = ffmpeg.file_info(input_path)

            if len(input_list) > 1:
                log.conwrite(f'Working on {inp.basename}')

            cuts, output_path = edit_media(i, inp, ffmpeg, args, progress,
                segments[i], exporting_to_editor, making_data_file, TEMP, log)
            num_cuts += cuts

        if not args.preview and not args.timeline and not making_data_file:
            timer.stop()

        if not args.preview and not args.timeline and making_data_file:
            # Assume making each cut takes about 30 seconds.
            time_save = usefulfunctions.human_readable_time(num_cuts * 30)
            s = 's' if num_cuts != 1 else ''

            log.print('Auto-Editor made {} cut{}, which would have taken about {} if '
                'edited manually.'.format(num_cuts, s, time_save))

        if not args.no_open:
            usefulfunctions.open_with_system_default(output_path, log)

    try:
        main_loop(input_list, ffmpeg, args, segments, log)
    except KeyboardInterrupt:
        log.error('Keyboard Interrupt')
    log.cleanup()

if __name__ == '__main__':
    main()



import sys

import av
import numpy as np
from PIL import Image, ImageOps, ImageChops, ImageFilter

from auto_editor.utils.progressbar import ProgressBar


def new_size(size, width):
    h, w = size
    r = width / w
    return width, int(h * r)


def display_motion_levels(path: str, width: int, blur: int):

    # TODO: Overwrite level when index is repeated. (Occurs in VFR videos)

    container = av.open(path, 'r')

    video_stream = container.streams.video[0]
    video_stream.thread_type = 'AUTO'

    prev_image = None
    image = None
    total_pixels = None

    for frame in container.decode(video_stream):
        if image is None:
            prev_image = None
        else:
            prev_image = image

        image = frame.to_image()

        if total_pixels is None:
            total_pixels = image.size[0] * image.size[1]

        image.thumbnail(new_size(image.size, width))
        image = ImageOps.grayscale(image)

        if blur > 0:
            image = image.filter(ImageFilter.GaussianBlur(radius=blur))

        if prev_image is not None:
            count = np.count_nonzero(ImageChops.difference(prev_image, image))

            sys.stdout.write('{:.20f}\n'.format(count))


def motion_detection(path: str, fps: float, threshold: float, progress: ProgressBar,
    width: int, blur: int) -> np.ndarray:

    container = av.open(path, 'r')

    video_stream = container.streams.video[0]
    video_stream.thread_type = 'AUTO'

    inaccurate_dur = int(float(video_stream.duration * video_stream.time_base) * fps)

    progress.start(inaccurate_dur, 'Analyzing motion')

    prev_image = None
    image = None
    total_pixels = None
    index = 0

    has_motion = np.zeros((1024), dtype=np.bool_)

    for frame in container.decode(video_stream):
        if image is None:
            prev_image = None
        else:
            prev_image = image

        index = int(frame.time * fps)

        progress.tick(index)

        if index > len(has_motion) - 1:
            has_motion = np.concatenate(
                (has_motion, np.zeros((len(has_motion)), dtype=np.bool_)), axis=0
            )

        image = frame.to_image()

        if total_pixels is None:
            total_pixels = image.size[0] * image.size[1]

        image.thumbnail(new_size(image.size, width))
        image = ImageOps.grayscale(image)

        if blur > 0:
            image = image.filter(ImageFilter.GaussianBlur(radius=blur))

        if prev_image is not None:
            count = np.count_nonzero(ImageChops.difference(prev_image, image))

            if count / total_pixels >= threshold:
                has_motion[index] = True

    progress.end()
    return has_motion[:index]



import os.path
from os.path import abspath
from shutil import move
from urllib.parse import quote
from platform import system

from .utils import indent, get_width_height, safe_mkdir

PIXEL_ASPECT_RATIO = 'square'
COLORDEPTH = '24'
NTSC = 'FALSE'
ANAMORPHIC = 'FALSE'
DEPTH = '16'

def fix_url(path: str) -> str:
    if system() == 'Windows':
        return'file://localhost/' + quote(abspath(path)).replace('%5C', '/')
    return 'file://localhost' + abspath(path)


def speedup(speed: float) -> str:
    return indent(6, '<filter>', '\t<effect>', '\t\t<name>Time Remap</name>',
        '\t\t<effectid>timeremap</effectid>',
        '\t\t<effectcategory>motion</effectcategory>',
        '\t\t<effecttype>motion</effecttype>',
        '\t\t<mediatype>video</mediatype>',
        '\t\t<parameter authoringApp="PremierePro">',
        '\t\t\t<parameterid>variablespeed</parameterid>',
        '\t\t\t<name>variablespeed</name>', '\t\t\t<valuemin>0</valuemin>',
        '\t\t\t<valuemax>1</valuemax>',
        '\t\t\t<value>0</value>',
        '\t\t</parameter>',
        '\t\t<parameter authoringApp="PremierePro">',
        '\t\t\t<parameterid>speed</parameterid>',  '\t\t\t<name>speed</name>',
        '\t\t\t<valuemin>-100000</valuemin>', '\t\t\t<valuemax>100000</valuemax>',
        f'\t\t\t<value>{speed}</value>',
        '\t\t</parameter>',
        '\t\t<parameter authoringApp="PremierePro">',
        '\t\t\t<parameterid>reverse</parameterid>',
        '\t\t\t<name>reverse</name>', '\t\t\t<value>FALSE</value>',
        '\t\t</parameter>',
        '\t\t<parameter authoringApp="PremierePro">',
        '\t\t\t<parameterid>frameblending</parameterid>',
        '\t\t\t<name>frameblending</name>', '\t\t\t<value>FALSE</value>',
        '\t\t</parameter>', '\t</effect>', '</filter>')


def premiere_xml(inp, temp, output, chunks, sampleRate, fps, log):

    audio_file = len(inp.video_streams) == 0 and len(inp.audio_streams) == 1

    clips = list(filter(lambda chunk: chunk[2] != 99999, chunks))

    duration = chunks[-1][1]
    pathurls = [fix_url(inp.path)]

    tracks = len(inp.audio_streams)

    if tracks > 1:
        name_without_extension = inp.basename[:inp.basename.rfind('.')]

        fold = safe_mkdir(os.path.join(inp.dirname, '{}_tracks'.format(
            name_without_extension)))

        for i in range(1, tracks):
            newtrack = os.path.join(fold, f'{i}.wav')
            move(os.path.join(temp, f'{i}.wav'), newtrack)
            pathurls.append(fix_url(newtrack))

    width, height = get_width_height(inp)
    if width is None or height is None:
        width, height = '1280', '720'

    timebase = str(int(fps))

    group_name = 'Auto-Editor {} Group'.format('Audio' if audio_file else 'Video')

    with open(output, 'w', encoding='utf-8') as outfile:
        outfile.write('<?xml version="1.0" encoding="UTF-8"?>\n<!DOCTYPE xmeml>\n')
        outfile.write('<xmeml version="4">\n')
        outfile.write('\t<sequence>\n')
        outfile.write(f'\t\t<name>{group_name}</name>\n')
        outfile.write(f'\t\t<duration>{duration}</duration>\n')
        outfile.write('\t\t<rate>\n')
        outfile.write(f'\t\t\t<timebase>{timebase}</timebase>\n')
        outfile.write(f'\t\t\t<ntsc>{NTSC}</ntsc>\n')
        outfile.write('\t\t</rate>\n')
        outfile.write('\t\t<media>\n')
        outfile.write(indent(3,
            '<video>',
            '\t<format>',
            '\t\t<samplecharacteristics>',
            )
        )

        if len(inp.video_streams) > 0:
            outfile.write(indent(3,
                '\t\t\t<rate>',
                f'\t\t\t\t<timebase>{timebase}</timebase>',
                f'\t\t\t\t<ntsc>{NTSC}</ntsc>',
                '\t\t\t</rate>')
            )

        outfile.write(indent(3,
            '\t\t\t<width>{}</width>'.format(width),
            '\t\t\t<height>{}</height>'.format(height),
            f'\t\t\t<pixelaspectratio>{PIXEL_ASPECT_RATIO}</pixelaspectratio>',
        ))

        if len(inp.video_streams) > 0:
            outfile.write(indent(3,
                '\t\t\t<fielddominance>none</fielddominance>',
                f'\t\t\t<colordepth>{COLORDEPTH}</colordepth>',
                )
            )

        outfile.write(indent(3,
            '\t\t</samplecharacteristics>',
            '\t</format>',
            '</video>' if len(inp.video_streams) == 0 else '\t<track>'
            )
        )

        if len(inp.video_streams) > 0:
            # Handle video clips

            total = 0
            for j, clip in enumerate(clips):

                clip_duration = (clip[1] - clip[0]) / clip[2]

                my_start = int(total)
                my_end = int(total) + int(clip_duration)

                total += clip_duration

                outfile.write(indent(5,
                    '<clipitem id="clipitem-{}">'.format(j+1),
                    '\t<masterclipid>masterclip-2</masterclipid>',
                    '\t<name>{}</name>'.format(inp.basename),
                    '\t<start>{}</start>'.format(my_start),
                    '\t<end>{}</end>'.format(my_end),
                    '\t<in>{}</in>'.format(int(clip[0] / clip[2])),
                    '\t<out>{}</out>'.format(int(clip[1] / clip[2]))
                    )
                )

                if j == 0:
                    outfile.write(indent(6,
                        '<file id="file-1">',
                        '\t<name>{}</name>'.format(inp.basename),
                        '\t<pathurl>{}</pathurl>'.format(pathurls[0]),
                        '\t<rate>',
                        f'\t\t<timebase>{timebase}</timebase>',
                        f'\t\t<ntsc>{NTSC}</ntsc>',
                        '\t</rate>',
                        f'\t<duration>{duration}</duration>',
                        '\t<media>',
                        '\t\t<video>',
                        '\t\t\t<samplecharacteristics>',
                        '\t\t\t\t<rate>',
                        f'\t\t\t\t\t<timebase>{timebase}</timebase>',
                        f'\t\t\t\t\t<ntsc>{NTSC}</ntsc>',
                        '\t\t\t\t</rate>',
                        '\t\t\t\t<width>{}</width>'.format(width),
                        '\t\t\t\t<height>{}</height>'.format(height),
                        f'\t\t\t\t<anamorphic>{ANAMORPHIC}</anamorphic>',
                        f'\t\t\t\t<pixelaspectratio>{PIXEL_ASPECT_RATIO}</pixelaspectratio>',
                        '\t\t\t\t<fielddominance>none</fielddominance>',
                        '\t\t\t</samplecharacteristics>',
                        '\t\t</video>',
                        '\t\t<audio>',
                        '\t\t\t<samplecharacteristics>',
                        f'\t\t\t\t<depth>{DEPTH}</depth>',
                        f'\t\t\t\t<samplerate>{sampleRate}</samplerate>',
                        '\t\t\t</samplecharacteristics>',
                        '\t\t\t<channelcount>2</channelcount>',
                        '\t\t</audio>',
                        '\t</media>',
                        '</file>',
                        )
                    )
                else:
                    outfile.write('\t\t\t\t\t\t<file id="file-1"/>\n')

                if(clip[2] != 1):
                    outfile.write(speedup(clip[2] * 100))

                # Linking for video blocks
                for i in range(max(3, tracks + 1)):
                    outfile.write('\t\t\t\t\t\t<link>\n')
                    outfile.write('\t\t\t\t\t\t\t<linkclipref>clipitem-{}</linkclipref>\n'.format((i*(len(clips)))+j+1))
                    if(i == 0):
                        outfile.write('\t\t\t\t\t\t\t<mediatype>video</mediatype>\n')
                    else:
                        outfile.write('\t\t\t\t\t\t\t<mediatype>audio</mediatype>\n')
                    if(i == 2):
                        outfile.write('\t\t\t\t\t\t\t<trackindex>2</trackindex>\n')
                    else:
                        outfile.write('\t\t\t\t\t\t\t<trackindex>1</trackindex>\n')
                    outfile.write('\t\t\t\t\t\t\t<clipindex>{}</clipindex>\n'.format(j+1))
                    if(i > 0):
                        outfile.write('\t\t\t\t\t\t\t<groupindex>1</groupindex>\n')
                    outfile.write('\t\t\t\t\t\t</link>\n')

                outfile.write('\t\t\t\t\t</clipitem>\n')
            outfile.write(indent(3, '\t</track>', '</video>'))


        # Audio Clips
        outfile.write(indent(3,
            '<audio>',
            '\t<numOutputChannels>2</numOutputChannels>',
            '\t<format>',
            '\t\t<samplecharacteristics>',
            f'\t\t\t<depth>{DEPTH}</depth>',
            f'\t\t\t<samplerate>{sampleRate}</samplerate>',
            '\t\t</samplecharacteristics>',
            '\t</format>'
            )
        )

        for t in range(tracks):
            outfile.write('\t\t\t\t<track currentExplodedTrackIndex="0" premiereTrackType="Stereo">\n')
            total = 0
            for j, clip in enumerate(clips):

                clip_duration = (clip[1] - clip[0]) / clip[2]

                my_start = int(total)
                my_end = int(total) + int(clip_duration)

                total += clip_duration

                if audio_file:
                    clip_item_num = j + 1
                    master_id = '1'
                else:
                    clip_item_num = len(clips) + 1 + j + (t * len(clips))
                    master_id = '2'

                outfile.write(indent(5,
                    f'<clipitem id="clipitem-{clip_item_num}" premiereChannelType="stereo">',
                    f'\t<masterclipid>masterclip-{master_id}</masterclipid>',
                    f'\t<name>{inp.basename}</name>',
                    f'\t<start>{my_start}</start>',
                    f'\t<end>{my_end}</end>',
                    '\t<in>{}</in>'.format(int(clip[0] / clip[2])),
                    '\t<out>{}</out>'.format(int(clip[1] / clip[2]))
                    )
                )

                if j == 0 and (audio_file or t > 0):
                    outfile.write(indent(6,
                        '<file id="file-{}">'.format(t+1),
                        '\t<name>{}</name>'.format(inp.basename),
                        '\t<pathurl>{}</pathurl>'.format(pathurls[t]),
                        '\t<rate>',
                        f'\t\t<timebase>{timebase}</timebase>',
                        f'\t\t<ntsc>{NTSC}</ntsc>',
                        '\t</rate>',
                        '\t<media>',
                        '\t\t<audio>',
                        '\t\t\t<samplecharacteristics>',
                        f'\t\t\t\t<depth>{DEPTH}</depth>',
                        f'\t\t\t\t<samplerate>{sampleRate}</samplerate>',
                        '\t\t\t</samplecharacteristics>',
                        '\t\t\t<channelcount>2</channelcount>',
                        '\t\t</audio>',
                        '\t</media>',
                        '</file>',
                        )
                    )
                else:
                    outfile.write('\t\t\t\t\t\t<file id="file-{}"/>\n'.format(t+1))

                outfile.write(indent(6,
                    '<sourcetrack>',
                    '\t<mediatype>audio</mediatype>',
                    '\t<trackindex>1</trackindex>',
                    '</sourcetrack>'
                    '<labels>',
                    '\t<label2>Iris</label2>',
                    '</labels>'
                    )
                )

                if clip[2] != 1:
                    outfile.write(speedup(clip[2] * 100))

                outfile.write('\t\t\t\t\t</clipitem>\n')
            if not audio_file:
                outfile.write('\t\t\t\t\t<outputchannelindex>1</outputchannelindex>\n')
            outfile.write('\t\t\t\t</track>\n')

        outfile.write('\t\t\t</audio>\n')
        outfile.write('\t\t</media>\n')
        outfile.write('\t</sequence>\n')
        outfile.write('</xmeml>\n')

    log.conwrite('')



"""
To prevent duplicate code being pasted between scripts, common functions should be
put here. No function should modify or create video/audio files on its own.
"""

from typing import List, Tuple, Union
from auto_editor.utils.log import Log

def get_stdout(cmd: List[str]) -> str:
    from subprocess import Popen, PIPE, STDOUT

    stdout, _ = Popen(cmd, stdout=PIPE, stderr=STDOUT).communicate()
    return stdout.decode()

def clean_list(x: List[str], rm_chars: str) -> list:
    new_list = []
    for item in x:
        for char in rm_chars:
            item = item.replace(char, '')
        new_list.append(item)
    return new_list

def aspect_ratio(width: int, height: int) -> Union[Tuple[int, int], Tuple[None, None]]:
    if height == 0:
        return None, None

    def gcd(a, b):
        while b:
            a, b = b, a % b
        return a

    c = gcd(width, height)
    return width // c, height // c

def get_new_length(chunks: List[Tuple[int, int, float]], fps: float) -> float:
    time_in_frames = 0.0
    for chunk in chunks:
        leng = chunk[1] - chunk[0]
        if chunk[2] != 99999:
            time_in_frames += leng * (1 / chunk[2])
    return time_in_frames / fps

def human_readable_time(time_in_secs: Union[int, float]) -> str:
    units = 'seconds'
    if time_in_secs >= 3600:
        time_in_secs = round(time_in_secs / 3600, 1)
        if time_in_secs % 1 == 0:
            time_in_secs = round(time_in_secs)
        units = 'hours'
    if time_in_secs >= 60:
        time_in_secs = round(time_in_secs / 60, 1)
        if time_in_secs >= 10 or time_in_secs % 1 == 0:
            time_in_secs = round(time_in_secs)
        units = 'minutes'
    return '{} {}'.format(time_in_secs, units)

def open_with_system_default(path: str, log: Log) -> None:
    from subprocess import call
    try:  # should work on Windows
        from os import startfile
        startfile(path)
    except ImportError:
        try:  # should work on MacOS and most Linux versions
            call(['open', path])
        except Exception:
            try: # should work on WSL2
                call(['cmd.exe', '/C', 'start', path])
            except Exception:
                try: # should work on various other Linux distros
                    call(['xdg-open', path])
                except Exception:
                    log.warning('Could not open output file.')

def fnone(val: object) -> bool:
    return val == 'none' or val == 'unset' or val is None

def append_filename(path: str, val: str) -> str:
    from os.path import splitext

    root, ext = splitext(path)
    return root + val + ext

def set_output_name(path: str, inp_ext: str, making_data_file: bool, args: object) -> str:
    from os.path import splitext

    root, ext = splitext(path)

    if args.export_as_json:
        return root + '.json'
    if args.export_to_final_cut_pro:
        return root + '.fcpxml'
    if args.export_to_shotcut:
        return root + '.mlt'
    if making_data_file:
        return root + '.xml'
    if args.export_as_audio:
        return root + '_ALTERED.wav'
    if ext == '':
        if inp_ext is None:
            return root
        return root + inp_ext

    return root + '_ALTERED' + ext



# Internal Libraries
import os.path
import subprocess

# Included Libraries
from auto_editor.utils.encoder import encoders
from auto_editor.utils.video import get_vcodec, video_quality

# From: github.com/PyAV-Org/PyAV/blob/main/av/video/frame.pyx
allowed_pix_fmt = {'yuv420p', 'yuvj420p', 'rgb24', 'bgr24', 'argb', 'rgba',
    'abgr', 'bgra', 'gray', 'gray8', 'rgb8', 'bgr8', 'pal8'}

def pix_fmt_allowed(pix_fmt: str) -> bool:
    return pix_fmt in allowed_pix_fmt


def apply_anchor(x, y, width, height, anchor):
    if anchor == 'ce':
        x = int((x * 2 - width) / 2)
        y = int((y * 2 - height) / 2)
    if anchor == 'tr':
        x -= width
    if anchor == 'bl':
        y -= height
    if anchor == 'br':
        x -= width
        y -= height
    # Pillow uses 'tl' by default
    return x, y

def one_pos_two_pos(x, y, width, height, anchor):
    """Convert: x, y, width, height -> x1, y1, x2, y2"""

    if anchor == 'ce':
        x1 = x - int(width / 2)
        x2 = x + int(width / 2)
        y1 = y - int(height / 2)
        y2 = y + int(height / 2)

        return x1, y1, x2, y2

    if anchor in ('tr', 'br'):
        x1 = x - width
        x2 = x
    else:
        x1 = x
        x2 = x + width

    if anchor in ('tl', 'tr'):
        y1 = y
        y2 = y + height
    else:
        y1 = y
        y2 = y - height

    return x1, y1, x2, y2


def set_static_assets(all_objects, log):
    """Save reloading the same thing over and over."""

    new_objects = []

    if len(all_objects) > 0:
        try:
            from PIL import Image, ImageDraw, ImageFont, ImageChops
        except ImportError:
            log.import_error('Pillow')

    for obj in all_objects:
        if obj._type == 'text':
            try:
                obj.font = ImageFont.truetype(obj.font, obj.size)
            except OSError:
                if obj.font == 'default':
                    obj.font = ImageFont.load_default()
                else:
                    log.error(f"Font '{obj.font}' not found.")

        if obj._type == 'image':
            source = Image.open(obj.src)
            source = source.convert('RGBA')
            source = source.rotate(obj.rotate, expand=True)
            source = ImageChops.multiply(source,
                Image.new('RGBA', source.size,
                    (255, 255, 255, int(obj.opacity * 255))
                )
            )
            obj.src = source

        new_objects.append(obj)

    return new_objects


def render_objects(sheet, all_objects, index, frame, pix_fmt):
    from PIL import Image, ImageDraw

    img = frame.to_image().convert('RGBA')

    for item in sheet[index]:
        obj = all_objects[item]

        obj_img = Image.new('RGBA', img.size, (255, 255, 255, 0))
        draw = ImageDraw.Draw(obj_img)

        if obj._type == 'text':
            text_w, text_h = draw.textsize(obj.content, font=obj.font)
            pos = apply_anchor(obj.x, obj.y, text_w, text_h, 'ce')
            draw.text(pos, obj.content, font=obj.font, fill=obj.fill, align=obj.align,
                stroke_width=obj.stroke, stroke_fill=obj.strokecolor)

        if obj._type == 'rectangle':
            draw.rectangle(
                one_pos_two_pos(obj.x, obj.y, obj.width, obj.height, obj.anchor),
                fill=obj.fill, width=obj.stroke, outline=obj.strokecolor
            )

        if obj._type == 'ellipse':
            draw.ellipse(
                one_pos_two_pos(obj.x, obj.y, obj.width, obj.height, obj.anchor),
                fill=obj.fill, width=obj.stroke, outline=obj.strokecolor
            )

        if obj._type == 'image':
            img_w, img_h = obj.src.size
            pos = apply_anchor(obj.x, obj.y, img_w, img_h, obj.anchor)
            obj_img.paste(obj.src, pos)

        img = Image.alpha_composite(img, obj_img)

    return frame.from_image(img).reformat(format=pix_fmt)


def render_av(ffmpeg, track, inp, args, chunks, fps, progress, effects, rules, temp, log):
    try:
        import av
    except ImportError:
        log.import_error('av')

    if chunks[-1][2] == 99999:
        chunks.pop()

    total_frames = chunks[-1][1]
    progress.start(total_frames, 'Creating new video')

    container = av.open(inp.path, 'r')
    pix_fmt = container.streams.video[track].pix_fmt

    target_pix_fmt = pix_fmt

    if not pix_fmt_allowed(pix_fmt):
        target_pix_fmt = 'yuv420p'

    my_codec = get_vcodec(args, inp, rules)

    apply_video_later = True

    if my_codec in encoders:
        apply_video_later = encoders[my_codec]['pix_fmt'].isdisjoint(allowed_pix_fmt)

    if args.scale != 1:
        apply_video_later = False

    log.debug(f'apply video quality settings now: {not apply_video_later}')

    video_stream = container.streams.video[track]
    video_stream.thread_type = 'AUTO'

    width = video_stream.width
    height = video_stream.height

    effects.all = set_static_assets(effects.all, log)

    spedup = os.path.join(temp, f'spedup{track}.mp4')

    cmd = ['-hide_banner', '-y', '-f', 'rawvideo', '-c:v', 'rawvideo',
        '-pix_fmt', target_pix_fmt, '-s', f'{width}*{height}', '-framerate', str(fps),
        '-i', '-', '-pix_fmt', target_pix_fmt]

    if apply_video_later:
        cmd.extend(['-c:v', 'mpeg4', '-qscale:v', '1'])
    else:
        cmd = video_quality(cmd, args, inp, rules)

    cmd.append(spedup)

    process2 = ffmpeg.Popen(cmd, stdin=subprocess.PIPE, stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL)

    input_equavalent = 0.0
    output_equavalent = 0
    chunk = chunks.pop(0)

    try:
        for frame in container.decode(video_stream):
            index = int(frame.time * fps)

            if index > chunk[1]:
                if chunks:
                    chunk = chunks.pop(0)
                else:
                    break

            if chunk[2] != 99999:
                input_equavalent += (1 / chunk[2])

            while input_equavalent > output_equavalent:
                if index in effects.sheet:
                    frame = render_objects(
                        effects.sheet, effects.all, index, frame, target_pix_fmt
                    )
                elif pix_fmt != target_pix_fmt:
                    frame = frame.reformat(format=target_pix_fmt)

                in_bytes = frame.to_ndarray().tobytes()
                process2.stdin.write(in_bytes)
                output_equavalent += 1

            progress.tick(index)

        progress.end()
        process2.stdin.close()
        process2.wait()
    except (OSError, BrokenPipeError):
        progress.end()
        ffmpeg.run_check_errors(cmd, log, True)
        log.error('FFmpeg Error!')

    # Unfortunately, scaling has to be a concrete step.
    if args.scale != 1:
        sped_input = os.path.join(temp, f'spedup{track}.mp4')
        spedup = os.path.join(temp, f'scale{track}.mp4')
        cmd = ['-i', sped_input, '-vf', 'scale=iw*{s}:ih*{s}'.format(s=args.scale),
            spedup]

        check_errors = ffmpeg.pipe(cmd)
        if('Error' in check_errors or 'failed' in check_errors):
            if('-allow_sw 1' in check_errors):
                cmd.insert(-1, '-allow_sw')
                cmd.insert(-1, '1')
            # Run again to show errors even if it might not work.
            ffmpeg.run(cmd)

    return 'video', spedup, apply_video_later




# Internal Libraries
import os
import sys
import shutil
import platform
import subprocess
from time import perf_counter

# External Libraries
import av
import numpy as np

# Typing
from typing import List, Tuple, Callable, NoReturn, Optional, Union

# Included Libraries
from auto_editor.utils.func import clean_list
import auto_editor.vanparse as vanparse

def test_options(parser):
    parser.add_argument('--only', '-n', nargs='*')
    parser.add_argument('--help', '-h', action='store_true',
        help='Print info about the program or an option and exit.')
    parser.add_required('category', nargs=1, choices=['cli', 'sub', 'api', 'unit', 'all'],
        help='Set what category of tests to run.')
    return parser


def av_inspect(path: str, *args):
    container = av.open(path, 'r')

    media = {}
    if len(container.streams.video) > 0:
        video = container.streams.video[0]
        media['fps'] = video.average_rate
        media['resolution'] = video.width, video.height
        media['vcodec'] = video.codec.name
        media['vlanguage'] = video.language

    if len(container.streams.audio) > 0:
        audio = container.streams.audio[0]
        media['samplerate'] = audio.rate
        media['acodec'] = audio.codec.name
        media['alanguage'] = audio.language

    for key, expected in args:
        if media[key] != expected:
            raise Exception(
                f"Media attribute '{key}' was '{media[key]}', expected '{expected}'."
            )


def pipe_to_console(cmd: List[str]) -> Tuple[int, str, str]:
    process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = process.communicate()
    return process.returncode, stdout.decode('utf-8'), stderr.decode('utf-8')


def cleanup(the_dir: str) -> None:
    for item in os.listdir(the_dir):
        item = os.path.join(the_dir, item)
        if ('_ALTERED' in item or item.endswith('.xml') or item.endswith('.json')
            or item.endswith('.fcpxml') or item.endswith('.mlt')):
            os.remove(item)
        if item.endswith('_tracks'):
            shutil.rmtree(item)


def clean_all() -> None:
    cleanup('resources')
    cleanup(os.getcwd())


def get_runner() -> list:
    if platform.system() == 'Windows':
        return ['py', '-m', 'auto_editor']
    return ['python3', '-m', 'auto_editor']


def run_program(cmd: List[str]) -> None:
    no_open = '.' in cmd[0]
    cmd = get_runner() + cmd

    if no_open:
        cmd += ['--no_open']

    returncode, stdout, stderr = pipe_to_console(cmd)
    if returncode > 0:
        raise Exception('{}\n{}\n'.format(stdout, stderr))


def check_for_error(cmd: List[str], match=None) -> None:
    returncode, stdout, stderr = pipe_to_console(get_runner() + cmd)
    if returncode > 0:
        if 'Error!' in stderr:
            if match is not None and match not in stderr:
                raise Exception('Could\'t find "{}"'.format(match))
        else:
            raise Exception('Program crashed.\n{}\n{}'.format(stdout, stderr))
    else:
        raise Exception('Program should not respond with a code 0.')


def make_np_list(in_file: str, compare_file: str, the_speed: float) -> None:
    from auto_editor.scipy.wavfile import read
    from auto_editor.audiotsm2 import phasevocoder
    from auto_editor.audiotsm2.io.array import ArrReader, ArrWriter

    samplerate, sped_chunk = read(in_file)
    spedup_audio = np.zeros((0, 2), dtype=np.int16)
    channels = 2

    with ArrReader(sped_chunk, channels, samplerate, 2) as reader:
        with ArrWriter(spedup_audio, channels, samplerate, 2) as writer:
            phasevocoder(reader.channels, speed=the_speed).run(
                reader, writer
            )
            spedup_audio = writer.output

    loaded = np.load(compare_file)

    if not np.array_equal(spedup_audio, loaded['a']):
        if spedup_audio.shape == loaded['a'].shape:
            print(f'Both shapes ({spedup_audio.shape}) are same')
        else:
            print(spedup_audio.shape)
            print(loaded['a'].shape)

        result = np.subtract(spedup_audio, loaded['a'])

        print('result non-zero: {}'.format(np.count_nonzero(result)))
        print('len of spedup_audio: {}'.format(len(spedup_audio)))

        print(np.count_nonzero(result) / spedup_audio.shape[0], 'difference between arrays')

        raise Exception("file {} doesn't match array.".format(compare_file))

    # np.savez_compressed(out_file, a=spedup_audio)


class Tester():
    def __init__(self, args):
        self.passed_tests = 0
        self.failed_tests = 0
        self.args = args

    def run_test(self, func: Callable, cleanup=None, allow_fail=False) -> None:
        if self.args.only != [] and func.__name__ not in self.args.only:
            return

        start = perf_counter()
        try:
            func()
            end = perf_counter() - start
        except KeyboardInterrupt:
            print(f'Testing Interrupted by User.')
            clean_all()
            sys.exit(1)
        except Exception as e:
            self.failed_tests += 1
            print(f"Test '{func.__name__}' failed.")
            print(e)
            if not allow_fail:
                clean_all()
                sys.exit(1)
        else:
            self.passed_tests += 1
            print(f"Test '{func.__name__}' passed: {round(end, 2)} secs")
            if cleanup is not None:
                cleanup()

    def end(self) -> NoReturn:
        print('{}/{}'.format(self.passed_tests, self.passed_tests + self.failed_tests))
        clean_all()
        sys.exit(0)


def main(sys_args: Optional[List[str]]=None):
    parser = vanparse.ArgumentParser('test', 'version')
    parser = test_options(parser)

    if sys_args is None:
        sys_args = sys.argv[1:]

    try:
        args = parser.parse_args(sys_args)
    except vanparse.ParserError as e:
        print(e)
        sys.exit(1)


    ### Tests ###

    def help_tests():
        """check the help option, its short, and help on options and groups."""
        run_program(['--help'])
        run_program(['-h'])
        run_program(['--frame_margin', '--help'])
        run_program(['--frame_margin', '-h'])
        run_program(['--help', '--help'])
        run_program(['-h', '--help'])
        run_program(['--help', '-h'])
        run_program(['-h', '--help'])


    def version_test():
        """Test version flags and debug by itself."""
        run_program(['--version'])
        run_program(['-v'])
        run_program(['-V'])
        run_program(['--debug'])


    def parser_test():
        check_for_error(['example.mp4', '--block'], 'needs argument')


    def subtitle_tests():
        from auto_editor.render.subtitle import SubtitleParser
        test = SubtitleParser()
        test.contents = [
            [0, 10, "A"],
            [10, 20, "B"],
            [20, 30, "C"],
            [30, 40, "D"],
            [40, 50, "E"],
            [50, 60, "F"],
        ]
        chunks = [
            (0, 10, 1),
            (10, 20, 99999),
            (20, 30, 1),
            (30, 40, 99999),
            (40, 50, 1),
            (50, 60, 99999),
        ]
        test.edit(chunks)

        if test.contents != [[0, 10, "A"], [10, 20, "C"], [20, 30, "E"]]:
            raise ValueError('Incorrect subtitle results.')


    def tsm_1a5_test():
        make_np_list('resources/example_cut_s16le.wav',
            'resources/example_1.5_speed.npz', 1.5)


    def tsm_0a5_test():
        make_np_list('resources/example_cut_s16le.wav',
            'resources/example_0.5_speed.npz', 0.5)


    def tsm_2a0_test():
        make_np_list('resources/example_cut_s16le.wav',
            'resources/example_2.0_speed.npz', 2)


    def info():
        run_program(['info', 'example.mp4'])
        run_program(['info', 'resources/man_on_green_screen.mp4'])
        run_program(['info', 'resources/multi-track.mov'])
        run_program(['info', 'resources/newCommentary.mp3'])
        run_program(['info', 'resources/test.mkv'])


    def levels():
        run_program(['levels', 'resources/multi-track.mov'])
        run_program(['levels', 'resources/newCommentary.mp3'])


    def subdump():
        run_program(['subdump', 'resources/subtitle.mp4'])


    def grep():
        run_program(['grep', 'boop', 'resources/subtitle.mp4'])


    def desc():
        run_program(['desc', 'example.mp4'])


    def example_tests():
        run_program(['example.mp4', '--video_codec', 'uncompressed'])
        av_inspect('example_ALTERED.mp4',
            ('fps', 30),
            ('resolution', (1280, 720), ),
            ('vcodec', 'mpeg4'),
            ('acodec', 'aac'),
            ('samplerate', 48000),
        )

        run_program(['example.mp4'])
        av_inspect('example_ALTERED.mp4',
            ('fps', 30),
            ('resolution', (1280, 720), ),
            ('vcodec', 'h264'),
            ('acodec', 'aac'),
            ('samplerate', 48000),
            ('vlanguage', 'eng'),
            ('alanguage', 'eng'),
        )

    # Issue #200
    def url_test():
        run_program(['https://github.com/WyattBlue/auto-editor/raw/master/example.mp4'])


    # Issue #184
    def unit_tests():
        """
        Make sure all units are working appropriately. That includes:

         - Seconds units: s, sec, secs, second, seconds
         - Frame units:   f, frame, frames
         - Sample units:  Hz, kHz
         - Percent:       %
        """

        run_program(['example.mp4', '--mark_as_loud', '20s,22sec', '25secs,26.5seconds'])
        run_program(['example.mp4', '--sample_rate', '44100'])
        run_program(['example.mp4', '--sample_rate', '44100 Hz'])
        run_program(['example.mp4', '--sample_rate', '44.1 kHz'])
        run_program(['example.mp4', '--silent_threshold', '4%'])


    def backwards_range_test():
        """
        Cut out the last 5 seconds of a media file by using negative number in the
        range.
        """
        run_program(['example.mp4', '--edit', 'none', '--cut_out', '-5secs,end'])
        run_program(['example.mp4', '--edit', 'all', '--add_in', '-5secs,end'])


    def cut_out_test():
        run_program(['example.mp4', '--edit', 'none', '--video_speed', '2', '--silent_speed',
            '3', '--cut_out', '2secs,10secs'])
        run_program(['example.mp4', '--edit', 'all', '--video_speed', '2', '--add_in',
            '2secs,10secs'])


    def gif_test():
        """
        Feed auto-editor a gif file and make sure it can spit out a correctly formated
        gif. No editing is requested.
        """
        run_program(['resources/man_on_green_screen.gif', '--edit', 'none'])
        av_inspect('resources/man_on_green_screen_ALTERED.gif',
            ('vcodec', 'gif'),
        )


    def margin_tests():
        run_program(['example.mp4', '-m', '3'])
        run_program(['example.mp4', '--margin', '3'])
        run_program(['example.mp4', '-m', '0.3sec'])
        run_program(['example.mp4', '-m', '6f,-3secs'])
        run_program(['example.mp4', '-m', '3,5 frames'])
        run_program(['example.mp4', '-m', '0.4 seconds'])


    def input_extension():
        """Input file must have an extension. Throw error if none is given."""

        shutil.copy('example.mp4', 'example')
        check_for_error(['example', '--no_open'], 'must have an extension.')
        os.remove('example')


    def output_extension():
        # Add input extension to output name if no output extension is given.
        run_program(['example.mp4', '-o', 'out'])
        av_inspect('out.mp4',
            ('vcodec', 'h264'),
        )
        os.remove('out.mp4')

        run_program(['resources/test.mkv', '-o', 'out'])
        av_inspect('out.mkv',
            ('vcodec', 'h264'),
        )
        os.remove('out.mkv')


    def progress_ops_test():
        run_program(['example.mp4', '--progress', 'machine'])
        run_program(['example.mp4', '--progress', 'none'])
        run_program(['example.mp4', '--progress', 'ascii'])


    def silent_threshold():
        run_program(['resources/newCommentary.mp3', '--silent_threshold', '0.1'])


    def track_tests():
        run_program(['resources/multi-track.mov', '--cut_by_all_tracks'])
        run_program(['resources/multi-track.mov', '--keep_tracks_seperate'])
        run_program(['example.mp4', '--cut_by_this_audio', 'resources/newCommentary.mp3'])


    def json_tests():
        run_program(['example.mp4', '--export_as_json'])
        run_program(['example.json'])


    def scale_tests():
        run_program(['example.mp4', '--scale', '1.5'])
        av_inspect('example_ALTERED.mp4',
            ('fps', 30),
            ('resolution', (1920, 1080), ),
            ('samplerate', 48000),
        )

        run_program(['example.mp4', '--scale', '0.2'])
        av_inspect('example_ALTERED.mp4',
            ('fps', 30),
            ('resolution', (256, 144), ),
            ('samplerate', 48000),
        )


    def various_errors_test():
        check_for_error(['example.mp4', '--add_rectangle', '0,60', '--cut_out', '60,end'])


    def effect_tests():
        """Test rendering video objects"""
        run_program(['create', 'test', '--width', '640', '--height', '360', '-o',
            'testsrc.mp4'])
        av_inspect('testsrc.mp4',
            ('fps', 30),
            ('resolution', (640, 360), ),
        )
        run_program(['testsrc.mp4', '--mark_as_loud', 'start,end', '--add_rectangle',
            '0,30,0,200,100,300,fill=#43FA56,stroke=10'])
        os.remove('testsrc_ALTERED.mp4')
        os.remove('testsrc.mp4')


    def render_text():
        run_program(['example.mp4', '--add-text', 'start,end,This is my text'])


    def check_font_error():
        check_for_error(
            ['example.mp4', '--add-text', 'start,end,text,0,0,30,notafont'], 'not found'
        )


    def export_tests():
        for test_name in ('aac.m4a', 'alac.m4a', 'pcm_f32le.wav', 'multi-track.mov',
            'pcm_s32le.wav', 'subtitle.mp4', 'test.mkv'):

            test_file = f'resources/{test_name}'
            run_program([test_file])
            run_program([test_file, '-exp'])
            run_program([test_file, '-exf'])
            run_program([test_file, '-exs'])
            run_program([test_file, '--export_as_clip_sequence'])
            run_program([test_file, '--preview'])
            cleanup('resources')


    def codec_tests():
        run_program(['example.mp4', '--video_codec', 'h264'])
        run_program(['example.mp4', '--audio_codec', 'ac3'])


    def combine_tests():
        run_program(['example.mp4', '--mark_as_silent', '0,171', '-o', 'hmm.mp4'])
        run_program(['example.mp4', 'hmm.mp4', '--combine_files', '--debug'])
        os.remove('hmm.mp4')


    def motion_tests():
        run_program(['resources/man_on_green_screen.mp4', '--edit_based_on', 'motion',
            '--debug', '--frame_margin', '0', '-mcut', '0', '-mclip', '0'])
        run_program(['resources/man_on_green_screen.mp4', '--edit_based_on', 'motion',
            '--motion_threshold', '0'])

    ### Runners ###

    tester = Tester(args)

    if args.category in ('unit', 'all'):
        tester.run_test(subtitle_tests)
        tester.run_test(tsm_1a5_test)
        tester.run_test(tsm_0a5_test, allow_fail=True)
        tester.run_test(tsm_2a0_test)

    if args.category in ('sub', 'all'):
        tester.run_test(info)
        tester.run_test(levels)
        tester.run_test(subdump)
        tester.run_test(grep)
        tester.run_test(desc)

    if args.category in ('cli', 'all'):
        tester.run_test(help_tests)
        tester.run_test(version_test)
        tester.run_test(parser_test)
        tester.run_test(example_tests, allow_fail=True)
        tester.run_test(url_test)
        tester.run_test(unit_tests)
        tester.run_test(backwards_range_test)
        tester.run_test(cut_out_test)
        tester.run_test(gif_test, cleanup=clean_all)
        tester.run_test(margin_tests)
        tester.run_test(input_extension)
        tester.run_test(output_extension)
        tester.run_test(progress_ops_test)
        tester.run_test(silent_threshold)
        tester.run_test(track_tests)
        tester.run_test(json_tests)
        tester.run_test(scale_tests)
        tester.run_test(various_errors_test)
        tester.run_test(effect_tests, cleanup=clean_all)
        tester.run_test(render_text)
        tester.run_test(check_font_error)
        tester.run_test(export_tests)
        tester.run_test(codec_tests)
        tester.run_test(combine_tests)
        tester.run_test(motion_tests)

    tester.end()

if __name__ == '__main__':
    main()



import numpy as np
from math import ceil

from auto_editor.utils.progressbar import ProgressBar

def get_max_volume(s: np.ndarray) -> float:
    return max(float(np.max(s)), -float(np.min(s)))


def display_audio_levels(read_track, fps: float):
    import sys

    from auto_editor.scipy.wavfile import read

    sample_rate, audio_samples = read(read_track)

    max_volume = get_max_volume(audio_samples)
    sample_count = audio_samples.shape[0]

    sample_rate_per_frame = sample_rate / fps
    audio_frame_count = ceil(sample_count / sample_rate_per_frame)

    for i in range(audio_frame_count):
        start = int(i * sample_rate_per_frame)
        end = min(int((i+1) * sample_rate_per_frame), sample_count)
        audiochunks = audio_samples[start:end]
        sys.stdout.write('{:.20f}\n'.format(get_max_volume(audiochunks) / max_volume))


def audio_detection(
    audio_samples: np.ndarray,
    sample_rate: int,
    silent_threshold: float,
    fps: float,
    progress: ProgressBar,
    ) -> np.ndarray:

    max_volume = get_max_volume(audio_samples)
    sample_count = audio_samples.shape[0]

    sample_rate_per_frame = sample_rate / fps
    audio_frame_count = ceil(sample_count / sample_rate_per_frame)

    progress.start(audio_frame_count, 'Analyzing audio volume')

    has_loud_audio = np.zeros((audio_frame_count), dtype=np.bool_)

    # Calculate when the audio is loud or silent.
    for i in range(audio_frame_count):

        if i % 500 == 0:
            progress.tick(i)

        start = int(i * sample_rate_per_frame)
        end = min(int((i+1) * sample_rate_per_frame), sample_count)

        if get_max_volume(audio_samples[start:end]) / max_volume > silent_threshold:
            has_loud_audio[i] = True

    progress.end()
    return has_loud_audio



# Internal Libraries
import os
from typing import List, Tuple

# External Libraries
import numpy as np

# Included Libraries
from auto_editor.sheet import Sheet
from auto_editor.utils.log import Log
from auto_editor.utils.func import fnone, set_output_name, append_filename

def get_chunks(inp, segment, fps, args, progress, log, audio_samples=None,
    sample_rate=None):
    from auto_editor.cutting import (combine_audio_motion, combine_segment, to_speed_list,
        set_range, chunkify, apply_mark_as, apply_margin, seconds_to_frames, cook)

    start_margin, end_margin = args.frame_margin

    start_margin = seconds_to_frames(start_margin, fps)
    end_margin = seconds_to_frames(end_margin, fps)
    min_clip = seconds_to_frames(args.min_clip_length, fps)
    min_cut = seconds_to_frames(args.min_cut_length, fps)

    def get_has_loud(inp, args, fps, audio_samples, sample_rate, log) -> np.ndarray:
        from auto_editor.analyze.generic import get_np_list

        if args.edit_based_on == 'none':
            return get_np_list(inp.path, audio_samples, sample_rate, fps, np.ones)
        if args.edit_based_on == 'all':
            return get_np_list(inp.path, audio_samples, sample_rate, fps, np.zeros)

        audio_list, motion_list = None, None

        if 'audio' in args.edit_based_on:
            from auto_editor.analyze.audio import audio_detection

            if audio_samples is None:
                audio_list = get_np_list(inp.path, audio_samples, sample_rate, fps, np.ones)
            else:
                audio_list = audio_detection(audio_samples, sample_rate,
                    args.silent_threshold, fps, progress)

        if 'motion' in args.edit_based_on:
            from auto_editor.analyze.motion import motion_detection

            if len(inp.video_streams) == 0:
                motion_list = get_np_list(inp.path, audio_samples, sample_rate, fps, np.ones)
            else:
                motion_list = motion_detection(inp.path, fps, args.motion_threshold,
                    progress, args.md_width, args.md_blur)

        if audio_list is not None and motion_list is not None:
            if len(audio_list) > len(motion_list):
                audio_list = audio_list[:len(motion_list)]
            elif len(motion_list) > len(audio_list):
                motion_list = motion_list[:len(audio_list)]

        return combine_audio_motion(audio_list, motion_list, args.edit_based_on, log)

    has_loud = get_has_loud(inp, args, fps, audio_samples, sample_rate, log)
    has_loud_length = len(has_loud)
    has_loud = apply_mark_as(has_loud, has_loud_length, fps, args, log)
    has_loud = cook(has_loud, min_clip, min_cut)
    has_loud = apply_margin(has_loud, has_loud_length, start_margin, end_margin)

    if segment is not None:
        has_loud = combine_segment(has_loud, segment, fps)

    # Remove small clips/cuts created by applying other rules.
    has_loud = cook(has_loud, min_clip, min_cut)

    speed_list = to_speed_list(has_loud, args.video_speed, args.silent_speed)

    if args.cut_out != []:
        speed_list = set_range(speed_list, args.cut_out, fps, 99999, log)

    if args.add_in != []:
        speed_list = set_range(speed_list, args.add_in, fps, args.video_speed, log)

    if args.set_speed_for_range != []:
        for item in args.set_speed_for_range:
            speed_list = set_range(speed_list, [item[1:]], fps, item[0], log)

    return chunkify(speed_list, has_loud_length)


def edit_media(i, inp, ffmpeg, args, progress, segment, exporting_to_editor, data_file,
    temp, log):
    from auto_editor.utils.container import get_rules

    chunks = None
    if inp.ext == '.json':
        from auto_editor.formats.timeline import read_json_timeline

        args.background, input_path, chunks = read_json_timeline(inp.path, log)
        inp = ffmpeg.file_info(input_path)

        output_path = set_output_name(inp.path, inp.ext, data_file, args)
    else:
        output_path = args.output_file[i]
        if not os.path.isdir(inp.path) and os.path.splitext(output_path)[1] == '':
            output_path = set_output_name(output_path, inp.ext, data_file, args)

    log.debug('{} -> {}'.format(inp.path, output_path))

    output_container = os.path.splitext(output_path)[1].replace('.', '')

    # Check if export options make sense.
    rules = get_rules(output_container)
    codec_error = "'{}' codec is not supported in '{}' container."

    if not fnone(args.sample_rate):
        if rules['samplerate'] is not None and args.sample_rate not in rules['samplerate']:
            log.error(
                "'{}' container only supports samplerates: {}".format(output_container,
                rules['samplerate'])
            )

    vcodec = args.video_codec
    if vcodec == 'uncompressed':
        vcodec = 'mpeg4'
    if vcodec == 'copy':
        vcodec = inp.video_streams[0]['codec']

    if vcodec != 'auto':
        if rules['vstrict'] and vcodec not in rules['vcodecs']:
            log.error(codec_error.format(vcodec, output_container))

        if vcodec in rules['disallow_v']:
            log.error(codec_error.format(vcodec, output_container))

    acodec = args.audio_codec
    if acodec == 'copy':
        acodec = inp.audio_streams[0]['codec']
        log.debug(f'Settings acodec to {acodec}')

    if acodec not in ('unset', 'auto'):
        if rules['astrict'] and acodec not in rules['acodecs']:
            log.error(codec_error.format(acodec, output_container))

        if acodec in rules['disallow_a']:
            log.error(codec_error.format(acodec, output_container))

    if args.keep_tracks_seperate and rules['max_audio_streams'] == 1:
        log.warning(f"'{container}' container doesn't support multiple audio tracks.")

    if (os.path.isfile(output_path) and inp.path != output_path and not args.preview
        and not args.timeline):
        log.debug('Removing already existing file: {}'.format(output_path))
        os.remove(output_path)

    audio_samples = None
    tracks = len(inp.audio_streams)

    fps = 30.0 if inp.fps is None else float(inp.fps)

    if fps < 1:
        log.error('{}: Frame rate cannot be below 1. fps: {}'.format(inp.basename, fps))

    # Extract subtitles in their native format.
    if len(inp.subtitle_streams) > 0:
        cmd = ['-i', inp.path, '-hide_banner']
        for s, sub in enumerate(inp.subtitle_streams):
            cmd.extend(['-map', '0:s:{}'.format(s)])
        for s, sub in enumerate(inp.subtitle_streams):
            cmd.extend([os.path.join(temp, '{}s.{}'.format(s, sub['ext']))])
        ffmpeg.run(cmd)

    sample_rate = None

    if args.cut_by_this_track >= tracks and args.cut_by_this_track != 0:
        message = "You choose a track that doesn't exist.\nThere "
        if tracks == 1:
            message += 'is only {} track.\n'.format(tracks)
        else:
            message += 'are only {} tracks.\n'.format(tracks)
        for t in range(tracks):
            message += ' Track {}\n'.format(t)
        log.error(message)

    # Split audio tracks into: 0.wav, 1.wav, etc.
    log.conwrite('Extracting audio')

    cmd = ['-i', inp.path, '-hide_banner']
    for t in range(tracks):
        cmd.extend(['-map', '0:a:{}'.format(t), '-ac', '2',
            os.path.join(temp, '{}.wav'.format(t))])
    ffmpeg.run(cmd)
    del cmd

    if tracks != 0:
        if args.cut_by_all_tracks:
            temp_file = os.path.join(temp, 'combined.wav')
            cmd = ['-i', inp.path, '-filter_complex',
                '[0:a]amix=inputs={}:duration=longest'.format(tracks), '-ac', '2',
                '-f', 'wav', temp_file]
            ffmpeg.run(cmd)
            del cmd
        else:
            temp_file = os.path.join(temp, '{}.wav'.format(args.cut_by_this_track))

        from auto_editor.scipy.wavfile import read
        sample_rate, audio_samples = read(temp_file)

    if chunks is None:
        chunks = get_chunks(inp, segment, fps, args, progress, log, audio_samples,
            sample_rate)

    if len(chunks) == 1 and chunks[0][2] == 99999:
        log.error('The entire media is cut!')

    def is_clip(chunk: Tuple[int, int, float]) -> bool:
        return chunk[2] != 99999

    def number_of_cuts(chunks: List[Tuple[int, int, float]]) -> int:
        return len(list(filter(is_clip, chunks)))

    num_cuts = number_of_cuts(chunks)


    pool = args.add_text + args.add_rectangle + args.add_ellipse + args.add_image

    obj_sheet = Sheet(pool, inp, chunks, log)

    if args.export_as_json:
        from auto_editor.formats.timeline import make_json_timeline
        make_json_timeline(inp.path, output_path, obj_sheet, chunks, fps, args.background,
            log)
        return num_cuts, output_path

    if args.timeline:
        from auto_editor.formats.timeline import make_json_timeline
        make_json_timeline(inp.path, 0, obj_sheet, chunks, fps, args.background, log)
        return num_cuts, None

    if args.preview:
        from auto_editor.preview import preview
        preview(inp, chunks, log)
        return num_cuts, None

    if args.export_to_premiere:
        from auto_editor.formats.premiere import premiere_xml
        premiere_xml(inp, temp, output_path, chunks, sample_rate, fps, log)
        return num_cuts, output_path

    if args.export_to_final_cut_pro:
        from auto_editor.formats.final_cut_pro import fcp_xml

        fcp_xml(inp, temp, output_path, chunks, fps, log)
        return num_cuts, output_path

    if args.export_to_shotcut:
        from auto_editor.formats.shotcut import shotcut_xml

        shotcut_xml(inp, temp, output_path, chunks, fps, log)
        return num_cuts, output_path

    def pad_chunk(
        chunk: Tuple[int, int, float], total_frames: int
    ) -> List[Tuple[int, int, float]]:

        start = []
        end = []
        if chunk[0] != 0:
            start.append((0, chunk[0], 99999.0))

        if chunk[1] != total_frames - 1:
            end.append((chunk[1], total_frames - 1, 99999.0))

        return start + [chunk] + end


    def make_media(inp, chunks: List[Tuple[int, int, float]], output_path: str) -> None:
        from auto_editor.utils.video import mux_quality_media

        if rules['allow_subtitle']:
            from auto_editor.render.subtitle import cut_subtitles
            cut_subtitles(ffmpeg, inp, chunks, fps, temp, log)

        if rules['allow_audio']:
            from auto_editor.render.audio import make_new_audio

            for t in range(tracks):
                temp_file = os.path.join(temp, '{}.wav'.format(t))
                new_file = os.path.join(temp, 'new{}.wav'.format(t))
                make_new_audio(temp_file, new_file, chunks, log, fps, progress)

                if not os.path.isfile(new_file):
                    log.bug('Audio file not created.')

        video_stuff = []

        if rules['allow_video']:
            from auto_editor.render.video import render_av
            for v, vid in enumerate(inp.video_streams):
                if vid['codec'] in ('png', 'jpeg'):
                    video_stuff.append(('image', None, None))
                else:
                    video_stuff.append(render_av(ffmpeg, v, inp, args, chunks, fps,
                        progress, obj_sheet, rules, temp, log))

        log.conwrite('Writing output file')

        mux_quality_media(
            ffmpeg, video_stuff, rules, output_path, output_container, args, inp, temp, log
        )
        if output_path is not None and not os.path.isfile(output_path):
            log.bug(f'The file {output_path} was not created.')

    if args.export_as_clip_sequence:
        total_frames = chunks[-1][1]
        clip_num = 0
        for chunk in chunks:
            if chunk[2] == 99999:
                continue
            make_media(inp, pad_chunk(chunk, total_frames),
                append_filename(output_path, f'-{clip_num}'))
            clip_num += 1
    else:
        make_media(inp, chunks, output_path)
    return num_cuts, output_path



from math import ceil

import av
import numpy as np

def get_np_list(path, audio_samples, sample_rate, fps, func):
    if audio_samples is not None:
        sample_count = audio_samples.shape[0]
        sample_rate_per_frame = sample_rate / fps
        audio_frame_count = ceil(sample_count / sample_rate_per_frame)
        return func((audio_frame_count), dtype=np.bool_)

    video = av.open(path, 'r').streams.video[0]
    total_frames = int(float(video.duration * video.time_base) * fps)
    return func((total_frames), dtype=np.bool_)



import re
import sys
import difflib
import textwrap
from shutil import get_terminal_size

from typing import List, Optional


class ParserError(Exception):
    pass


def indent(text: str, prefix, predicate=None) -> str:
    if predicate is None:
        def predicate(line):
            return line.strip()

    def prefixed_lines():
        for line in text.splitlines(True):
            yield (prefix + line if predicate(line) else line)
    return ''.join(prefixed_lines())


def out(text: str) -> None:
    width = get_terminal_size().columns - 3

    indent_regex = re.compile(r'^(\s+)')
    wrapped_lines = []

    for line in text.split('\n'):
        exist_indent = re.search(indent_regex, line)
        pre_indent = exist_indent.groups()[0] if exist_indent else ''

        wrapped_lines.append(
            textwrap.fill(line, width=width, subsequent_indent=pre_indent)
        )

    print('\n'.join(wrapped_lines))


def print_option_help(option):
    from dataclasses import fields, _MISSING_TYPE

    text = '  ' + ', '.join(option['names']) + '\n    ' + option['help'] + '\n\n'
    if option['dataclass'] is not None:
        text += '    Arguments:\n    '

        args = []
        for field in fields(option['dataclass']):
            if field.name != '_type':
                arg = '{' + field.name
                if not isinstance(field.default, _MISSING_TYPE):
                    arg += '=' + str(field.default)
                arg += '}'
                args.append(arg)

        text += ','.join(args)

    if option['manual'] != '':
        text += '{}\n\n'.format(indent(option['manual'], '    '))

    if option['dataclass'] is not None:
        pass
    elif option['action'] == 'default':
        text += f"    type: {option['type'].__name__}\n"
        text += f"    nargs: {option['nargs']}\n"
        text += f"    default: {option['default']}\n"

        if option['range'] is not None:
            text += f"    range: {option['range']}\n"

        if option['choices'] is not None:
            text += '    choices: ' +  ', '.join(option['choices']) + '\n'
    elif option['action'] in ('store_true', 'store_false'):
        text += '    type: flag\n'
    else:
        text += '    type: unknown\n'

    out(text)


def print_program_help(options):
    text = ''
    for option in options:
        if option['_type'] == 'text':
            text += '\n  ' + option['text'] + '\n'
        elif option['_type'] == 'blank':
            text += '\n'
        elif (option['_type'] == 'required' or
            (option['_type'] == 'option' and not option['hidden'])):
            text += '  ' + ', '.join(option['names']) + ': ' + option['help'] + '\n'
    text += '\n'
    out(text)


def to_underscore(name: str) -> str:
    """Convert new style options to old style.  e.g. --hello-world -> --hello_world"""
    return name[:2] + name[2:].replace('-', '_')


def to_key(val: dict) -> str:
    """Convert option name to arg key.  e.g. --hello-world -> hello_world"""
    return val['names'][0][:2].replace('-', '') + val['names'][0][2:].replace('-', '_')


def get_option(name: str, options: List[dict]) -> Optional[dict]:
    for option in options:
        if name in option['names'] or name in map(to_underscore, option['names']):
            return option
    return None


class ArgumentParser:
    def __init__(self, program_name, version, description=None):
        self.program_name = program_name
        self._version = version
        self.description = description

        self.args = []
        self.kwarg_defaults = {
            'nargs': 1,
            'type': str,
            'default': None,
            'action': 'default',
            'range': None,
            'choices': None,
            'help': '',
            'dataclass': None,
            'hidden': False,
            'manual': '',
            '_type': 'option',
        }

        self.required_defaults = {
            'nargs': '*',
            'type': str,
            'choices': None,
            'help': '',
            '_type': 'required',
        }

    def add_argument(self, *args, **kwargs):
        my_dict = self.kwarg_defaults.copy()
        my_dict['names'] = args

        for key, val in kwargs.items():
            my_dict[key] = val

        self.args.append(my_dict)


    def add_required(self, *args, **kwargs):
        my_dict = self.required_defaults.copy()
        my_dict['names'] = args

        for key, val in kwargs.items():
            my_dict[key] = val

        self.args.append(my_dict)


    def add_text(self, text: str):
        self.args.append({'text': text, '_type': 'text'});


    def add_blank(self):
        self.args.append({'_type': 'blank'});


    def parse_args(self, sys_args: list):
        if sys_args == [] and self.description:
            out(self.description)
            sys.exit()

        if sys_args == ['-v'] or sys_args == ['-V']:
            out('{} version {}'.format(self.program_name, self._version))
            sys.exit()

        return ParseOptions(sys_args, self.args)


class ParseOptions:

    @staticmethod
    def parse_dataclass(unsplit_arguments, op):
        """
        Positional Arguments
            --rectangle 0,end,10,20,20,30,#000, ...

        Keyword Arguments
            --rectangle start=0,end=end,x1=10, ...
        """

        from dataclasses import fields

        ARG_SEP = ','
        KEYWORD_SEP = '='

        d_name = op['dataclass'].__name__

        keys = []
        for field in fields(op['dataclass']):
            keys.append(field.name)

        kwargs = {}
        args = []

        allow_positional_args = True

        for i, arg in enumerate(unsplit_arguments.split(ARG_SEP)):
            if i+1 > len(keys):
                raise ParserError(f"{d_name} has too many arguments, starting "
                    f"with '{arg}'.")

            if KEYWORD_SEP in arg:
                allow_positional_args = False

                parameters = arg.split(KEYWORD_SEP)
                if len(parameters) > 2:
                    raise ParserError(f"{d_name} invalid syntax: '{arg}'.")
                key, val = parameters
                if key not in keys:
                    raise ParserError(f"{d_name} got an unexpected keyword '{key}'")

                kwargs[key] = val
            elif allow_positional_args:
                args.append(arg)
            else:
                raise ParserError(f'{d_name} positional argument follows keyword argument.')

        try:
            dataclass_instance = op['dataclass'](*args, **kwargs)
        except TypeError as err:
            err_list = [d_name] + str(err).split(' ')[1:]
            raise ParserError(' '.join(err_list))

        return dataclass_instance


    @staticmethod
    def parse_arg(option: dict, arg):

        option_name = option['names'][0]

        if arg is None and option['nargs'] == 1:
            raise ParserError(f"{option_name} needs argument.")

        try:
            value = option['type'](arg)
        except TypeError as e:
            raise ParserError(str(e))

        if option['choices'] is not None and value not in option['choices']:
            my_choices = ', '.join(option['choices'])

            raise ParserError(f'{value} is not a choice for {option_name}\n'
                f'choices are:\n  {my_choices}')

        return value


    def set_arg_list(self, option_list: str, my_list: list, list_type: type) -> None:
        if list_type is not None:
            setattr(self, option_list, list(map(list_type, my_list)))
        else:
            setattr(self, option_list, my_list)


    def __init__(self, sys_args: List[str], options_reqs: List[dict]) -> None:

        # Partition options and requireds.
        options = []
        requireds = []
        option_names = []

        for item in options_reqs:
            if item['_type'] == 'option':
                options.append(item)

                for name in item['names']:
                    option_names.append(name)

                if item['action'] == 'store_true':
                    value = False
                elif item['action'] == 'store_false':
                    value = True
                elif item['nargs'] != 1:
                    value = []
                elif item['default'] is None:
                    value = None
                else:
                    value = item['type'](item['default'])
                setattr(self, to_key(item), value)

            if item['_type'] == 'required':
                requireds.append(item)

        # Figure out command line options changed by user.
        used_options = []

        req_list = []
        req_list_name = requireds[0]['names'][0]
        req_list_type = requireds[0]['type']
        setting_req_list = requireds[0]['nargs'] != 1

        option_list = []
        op_list_name = None
        op_list_type = str
        setting_op_list = False

        i = 0
        while i < len(sys_args):
            arg = sys_args[i]
            option = get_option(arg, options)

            if option is None:
                if setting_op_list:
                    if used_options and used_options[-1]['dataclass'] is not None:
                        op_list_type = None
                        arg = self.parse_dataclass(arg, used_options[-1])

                    option_list.append(arg)

                elif requireds and not arg.startswith('--'):

                    if requireds[0]['nargs'] == 1:
                        setattr(self, req_list_name, self.parse_arg(requireds[0], arg))
                        requireds.pop()
                    else:
                        req_list.append(arg)
                else:
                    label = 'option' if arg.startswith('--') else 'short'

                    # 'Did you mean' message might appear that options need a comma.
                    if arg.replace(',', '') in option_names:
                        raise ParserError(f"Option '{arg}' has an unnecessary comma.")

                    close_matches = difflib.get_close_matches(arg, option_names)
                    if close_matches:
                        raise ParserError(
                            'Unknown {}: {}\n\n    Did you mean:\n        '.format(
                            label, arg) + ', '.join(close_matches)
                        )
                    raise ParserError(f'Unknown {label}: {arg}')
            else:
                if op_list_name is not None:
                    self.set_arg_list(op_list_name, option_list, op_list_type)

                if option in used_options:
                    raise ParserError(f"Cannot repeat option {option['names'][0]} twice.")

                used_options.append(option)

                setting_op_list = False
                option_list = []
                op_list_name = None

                key = to_key(option)

                next_arg = None if i == len(sys_args) - 1 else sys_args[i+1]
                if next_arg == '-h' or next_arg == '--help':
                    print_option_help(option)
                    sys.exit()

                if option['nargs'] != 1:
                    setting_op_list = True
                    op_list_name = key
                    op_list_type = option['type']
                elif option['action'] == 'store_true':
                    value = True
                elif option['action'] == 'store_false':
                    value = False
                else:
                    value = self.parse_arg(option, next_arg)
                    i += 1
                setattr(self, key, value)

            i += 1

        if setting_op_list:
            self.set_arg_list(op_list_name, option_list, op_list_type)

        if setting_req_list:
            self.set_arg_list(req_list_name, req_list, req_list_type)

        if self.help:
            print_program_help(options_reqs)
            sys.exit()



from typing import NoReturn

import sys
from shutil import rmtree, get_terminal_size
from time import perf_counter, sleep
from datetime import timedelta

class Timer:
    __slots__ = ('start_time', 'quiet')
    def __init__(self, quiet=False) -> None:
        self.start_time = perf_counter()
        self.quiet = quiet

    def stop(self) -> None:
        if not self.quiet:
            second_len = round(perf_counter() - self.start_time, 2)
            minute_len = timedelta(seconds=round(second_len))

            sys.stdout.write(f'Finished. took {second_len} seconds ({minute_len})\n')

class Log:
    __slots__ = ('is_debug', 'quiet', 'temp')
    def __init__(self, show_debug=False, quiet=False, temp=None) -> None:
        self.is_debug = show_debug
        self.quiet = quiet
        self.temp = temp

    def debug(self, message: str) -> None:
        if self.is_debug:
            self.conwrite('')
            sys.stderr.write(f'Debug: {message}\n')

    def cleanup(self) -> None:
        if self.temp is None:
            return
        try:
            rmtree(self.temp)
            self.debug('Removed Temp Directory.')
        except PermissionError:
            sleep(0.1)
            try:
                rmtree(self.temp)
                self.debug('Removed Temp Directory.')
            except Exception:
                self.debug('Failed to delete temp dir.')
        except FileNotFoundError:
            # that's ok, the folder we are trying to remove is already gone
            pass

    def conwrite(self, message: str):
        if not self.quiet:
            buffer = get_terminal_size().columns - len(message) - 3
            sys.stdout.write('  ' + message + ' ' * buffer + '\r')
            try:
                sys.stdout.flush()
            except AttributeError:
                pass

    def error(self, message: str) -> NoReturn:
        self.conwrite('')
        message = message.replace('\t', '    ')
        sys.stderr.write(f'Error! {message}\n')
        self.cleanup()

        from platform import system

        if system() == 'Linux':
            sys.exit(1)
        else:
            try:
                sys.exit(1)
            except SystemExit:
                import os
                os._exit(1)

    def bug(self, message: str, bug_type='bug report') -> NoReturn:
        self.error(
            f'{message}\n\nSomething went wrong!\nCreate a {bug_type} at:\n  '
            'https://github.com/WyattBlue/auto-editor/issues/\n'
        )

    def import_error(self, lib: str) -> NoReturn:
        self.error(f"Python module '{lib}' not installed. Run:  pip install {lib}")

    def warning(self, message: str) -> None:
        if not self.quiet:
            sys.stderr.write(f'Warning! {message}\n')

    def print(self, message: str) -> None:
        if not self.quiet:
            sys.stdout.write(f'{message}\n')
